#
# Author: MN

from yambopy import *
from netCDF4 import Dataset
import numpy as np
from yambopy.tools.string import marquee
from yambopy.units import I
import shutil
import os
from tqdm import tqdm
import scipy.fft
from yambopy.io.cubetools import write_cube
from yambopy.dbs.latticedb import YamboLatticeDB
import scipy
try:
    from pykdtree.kdtree import KDTree 
    ## pykdtree is much faster and is recommanded
    ## pip install pykdtree
    ## useful in Dmat computation
except ImportError as e:
    from scipy.spatial import KDTree
from yambopy.kpoints import build_ktree, find_kpt
from yambopy.tools.function_profiler import func_profile
from yambopy.tools.citations import citation

class YamboWFDB:
    """
    A class to load and manipulate wavefunctions from Yambo (ns.wf).

    Example usage:

    ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
        import numpy as np
        from yambopy.dbs import wfdb
        from yambopy.dbs.latticedb import YamboLatticeDB
        from netCDF4 import Dataset

        np.set_printoptions(suppress=True)

        abc = wfdb.YamboWFDB(path='.', bands_range = [5,10])

        abc.expand_fullBZ() ## expand wfcs to full BZ

        abc.write2cube(ik=2,ib=2) ## write electronic wfc to .cube for visualization

        print(abc.get_spin_projections(ik = 2,ib= 0)) ## get spin projection for SOC systems

        Dmat = abc.Dmat() ## compute phase matrices for SAVE symmetries
    ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    Attributes:
        path (str): Path to the directory containing the wavefunction files.
        filename (str): Name of the wavefunction file (default: 'ns.wf').
        wf (numpy.ndarray): Wavefunctions stored as a 5D array [nkpoints, nspin, nbands, nspinor, ngvect].
        gvecs (numpy.ndarray): G-vectors for the wavefunctions [nkpoints, ngvect, 3] (in reduced/crystal coordiantes).
        kpts_iBZ (numpy.ndarray): K-points in the irreducible Brillouin Zone (iBZ) in crystal coordinates.
        ngvecs (numpy.ndarray): Number of meaningful G-vectors for each wavefunction.
        fft_box (numpy.ndarray): Default FFT grid size for real-space conversion.
        nkpoints (int): Number of k-points in the iBZ.
        nspin (int): Number of spin components.
        nspinor (int): Number of spinor components.
        nbands (int): Number of bands.
        min_bnd (int): Minimum band index loaded.

    Methods:
        read(bands_range=[]): Read wavefunctions from the file.
        get_spin_projections(ik, ib, s_z=np.array([[1, 0], [0, -1]])): Compute spin projections for operator sz.
        get_iBZ_wf(ik): Get wavefunctions and G-vectors for a specific k-point.
        wfcG2r(ik, ib, grid=[]): Convert wavefunctions from G-space to real space.
        write2cube(ik, ib, grid=[]): Write wavefunctions to a cube file.
        get_iBZ_kpt(ik): Get the k-point in crystal coordinates.
        rotate_wfc(ik, isym): Rotate wavefunctions using a symmetry operation.
        apply_symm(kvec, wfc_k, gvecs_k, time_rev, sym_mat, frac_vec=np.array([0, 0, 0])): Apply symmetry to wavefunctions.
        to_real_space(wfc_tmp, gvec_tmp, grid=[]): Convert wavefunctions to real space.
    """

    def __init__(self, path=None, save='SAVE', filename='ns.wf', bands_range=[], latdb=None):
        """
        Initialize the YamboWFDB class.

        Args:
            path (str, optional): Path to the directory containing the wavefunction files. Defaults to the current directory.
            save (str, optional): Subdirectory containing the wavefunction files. Defaults to 'SAVE'.
            filename (str, optional): Name of the wavefunction file. Defaults to 'ns.wf'.
            bands_range (list, optional): Range of bands to load. Defaults to all bands.
        """
        if path is None:
            path = os.getcwd()
        self.path = os.path.join(path, save)
        self.filename = filename

        # Read wavefunctions
        self.read(bands_range=bands_range, latdb=latdb)

    def read(self, bands_range=[], latdb=None):
        """
        Read wavefunctions from the file.

        Args:
            bands_range (list, optional): Range of bands to load. Defaults to all bands.
            latdb : latticedb, if None (default), it will be created internally
        """
        path = self.path
        filename = self.filename

        # Open the ns.db1 database to get essential data
        try:
            ns_db1_fname = os.path.join(path, 'ns.db1')
            if latdb :
                if not hasattr(latdb,'ibz_kpoints'): latdb.expand_kpoints()
                self.ydb = latdb
            else :
                self.ydb = YamboLatticeDB.from_db_file(ns_db1_fname, Expand=True)
            ## total kpoints in full BZ
            self.nkBZ = len(self.ydb.symmetry_indexes)
            #
            # Read G-vectors and other data from ns_db1
            ns_db1 = Dataset(ns_db1_fname, 'r')
            lat_vec = self.ydb.lat.T
            lat_param = self.ydb.alat

            # K-points in iBZ (crystal units)
            self.kpts_iBZ = self.ydb.ibz_kpoints / lat_param[None, :]
            self.kpts_iBZ = self.kpts_iBZ @ lat_vec

            # K-points in BZ (crystal units)
            self.kBZ = self.ydb.iku_kpoints / lat_param[None, :]
            self.kBZ = self.kBZ @ lat_vec
            self.ktree = build_ktree(self.kBZ)

            # G-vectors in cartesian units
            G_vec = ns_db1['G-VECTORS'][...].data.T
            wfc_grid = ns_db1['WFC_GRID'][...].data
            wfc_grid = np.rint(wfc_grid).astype(int)

            # Number of G-vectors for each wavefunction
            igk = np.array(np.rint(ns_db1['WFC_NG'][...].data), dtype=int)
            self.ngvecs = igk

            # Convert G-vectors to crystal units
            G_vec = G_vec / lat_param[None, :]
            G_vec = np.array(np.rint(G_vec @ lat_vec), dtype=int)

            # Determine the default FFT grid size
            min_fft_idx = np.min(G_vec, axis=0)
            max_fft_idx = np.max(G_vec, axis=0)
            assert np.min(max_fft_idx) >= 0 and np.max(min_fft_idx) < 0, "Invalid G-vectors"
            self.fft_box = np.zeros(3, dtype=int)
            for i in range(3):
                self.fft_box[i] = max_fft_idx[i] - min_fft_idx[i] + 1

            # Read dimensions
            dimensions = ns_db1['DIMENSIONS'][...].data
            self.nkpoints = int(np.rint(dimensions[6]))  # Number of k-points in iBZ
            self.nspin = int(np.rint(dimensions[12]))    # Number of spin components
            self.nspinor = int(np.rint(dimensions[11]))  # Number of spinor components
            self.nbands = int(np.rint(dimensions[5]))    # Number of bands

            # Validate bands_range
            if len(bands_range) == 0:
                bands_range = [0, self.nbands]
            elif min(bands_range) < 0 or max(bands_range) > self.nbands:
                print("Warning: Invalid bands_range, loading all bands.")
                bands_range = [0, self.nbands]

            self.min_bnd = min(bands_range)
            self.nbands = max(bands_range) - self.min_bnd

            ns_db1.close()
        except Exception as e:
            raise IOError(f'Cannot read ns.db1 file: {e}')

        # Load wavefunctions
        wf = []
        for ik in tqdm(range(self.nkpoints), desc="Loading Wavefunctions"):
            for ispin in range(self.nspin):
                try:
                    fname = f"{filename}_fragments_{ispin * self.nkpoints + ik + 1}_1"
                    fname = os.path.join(path, fname)
                    database = Dataset(fname, 'r')
                    database_var_name = f'WF_COMPONENTS_@_SP_POL{ispin + 1}_K{ik + 1}_BAND_GRP_1'
                    aux = database.variables[database_var_name][self.min_bnd:self.min_bnd + self.nbands, ...].data
                    aux = aux[..., 0] + 1j * aux[..., 1]
                    aux[..., igk[ik]:] = 0  # Set invalid components to zero
                    wf.append(aux)
                    database.close()
                except Exception as e:
                    raise IOError(f'Could not read {fname}: {e}')

        self.ng = wf[0].shape[-1]  # Maximum number of wavefunction components
        self.wf = np.array(wf).reshape(self.nkpoints, self.nspin, self.nbands, self.nspinor, self.ng)
        # (nk, nspin,nbands,nspinor,ngvec)
        # Load G-vectors
        self.gvecs = np.zeros((self.nkpoints, self.ng, 3), dtype=int)
        for ik in tqdm(range(self.nkpoints), desc="Loading Miller Indices"):
            self.gvecs[ik, igk[ik]:, :] = 2147483646 * np.array([1, 1, 1])[None, :]  # Invalid G-vector marker
            self.gvecs[ik, :igk[ik], :] = G_vec[wfc_grid[ik][:igk[ik]] - 1, :]

    def __str__(self):
        """Return a string representation of the object."""
        lines = []
        lines.append(marquee(self.__class__.__name__))
        lines.append(f"nkpoints: {self.nkpoints:4d}")
        lines.append(f"nspin:    {self.nspin:4d}")
        lines.append(f"nbands:   [{self.min_bnd:4d}, {self.min_bnd + self.nbands:4d})")
        lines.append(f"Max wf components: {self.ng:4d}")
        return "\n".join(lines)

    def assert_k_inrange(self, ik):
        """Assert that the k-point index is valid."""
        assert 0 <= ik < self.nkpoints, "Invalid k-point index"

    def assert_bnd_range(self, ib):
        """Assert that the band index is valid."""
        assert 0 <= ib < self.nbands, "Invalid band index"
    
    def kptBZidx(self, kpts):
        """ 
        return the index of the kpoint or kpoints in the wavefunction db.
        The kpts must be in crystal (reduced) coordinates.
        """
        return find_kpt(self.ktree, kpts)

    def get_spin_projections(self, ik, ib=-1, s_z=np.array([[1, 0], [0, -1]])):
        """
        Compute the diagonal expectation values of the spin operator for electronic states.
        
        Args:
            ik (int): ibz K-point index.
            ib (int): Band index/array of indices. Default (-1) for all bands
            s_z (numpy.ndarray, optional): Spin operator. Defaults to the z-component.

        Returns:
            numpy.ndarray: Spin projection values.
        """
        self.assert_k_inrange(ik)

        assert self.nspin == 1 and self.nspinor == 2, "Spin projections are only useful for nspin=1 and nspinor=2"
        assert s_z.shape == (2, 2), "Spin operator must be a 2x2 matrix"

        if np.isscalar(ib):
            if ib < 0 : wfc_tmp = self.wf[ik, 0]
            else : wfc_tmp = self.wf[ik, 0][[ib]]
        else: wfc_tmp = self.wf[ik, 0][ib] 

        s_tmp = s_z[None,:,:]@wfc_tmp #'ij,bjg->big', s_z, wfc_tmp
        tmp_nb = s_tmp.shape[0]
        s_proj = s_tmp.reshape(tmp_nb,1,-1) @ wfc_tmp.reshape(tmp_nb,1,-1).transpose(0,2,1).conj()
        return s_proj.reshape(tmp_nb)

    def get_spin_m_e_BZ(self,s_z=np.array([[1, 0], [0, -1]])):
        """
        Compute the spin matrix elements (including off-diagonal 
        i.e, we compute < k,m| S_z | k,n>. 
        ---
        Note : get_spin_projections() is a sepecial case where it outputs diagonal terms
        ---
        Args:
            s_z (numpy.ndarray, optional): Spin operator. Defaults to the z-component of spin operator.

        < Sk,m| S_z | Sk,n> = <k,m| (U^\dagger S_z U) | k,m>
        if S is tim rev: < Sk,m| S_z | Sk,n> = <k,m| (U^\dagger S_z^* U) | k,m>^*
        This means, we donot need to rotate the wfcs. 
        instead we need to rotate the spin operator with su(2) matrices 
        ---
        ---
        Returns:
            numpy.ndarray: Spin operator (nk_bz, nbnd,nbnd)
        """
        #
        assert self.nspin == 1 and self.nspinor == 2, \
            "Spin operator is useful only for nspin=1 and nspinor=2"
        #
        kpt_idx = self.ydb.kpoints_indexes
        sym_idx = self.ydb.symmetry_indexes
        sym_mat = self.ydb.sym_car
        nsym = len(sym_mat)
        ## compute su(2) matrices for symmetries 
        su2_ops = np.zeros((nsym,2,2),dtype=self.wf.dtype)
        for isym in range(nsym):
            trev_tmp = (isym >= nsym / (1 + int(np.rint(self.ydb.time_rev))))
            su2_ops[isym] = su2_mat(sym_mat[isym],trev_tmp)
        #
        time_rev = (sym_idx >= nsym / (1 + int(np.rint(self.ydb.time_rev))))
        #
        su2_k = su2_ops[sym_idx]
        S_z_tmp = np.zeros((len(sym_idx),2,2),dtype=self.wf.dtype)
        S_z_tmp[...] = s_z[None,:,:]
        #
        # COnjugate the S_z operator in case of time reversal
        S_z_tmp[time_rev] = S_z_tmp[time_rev].conj()
        # rotate the S_z operator 
        S_z_tmp = S_z_tmp@su2_k
        S_z_tmp = su2_k.conj().transpose(0,2,1)@S_z_tmp
        #
        # Apply to wfc
        wfc_tmps = self.wf[:,0,...][kpt_idx]
        nk, nb, nspinor, ng = wfc_tmps.shape
        S_me = S_z_tmp[:,None,:,:]@wfc_tmps
        #
        ## compute the inner product
        S_me = wfc_tmps.reshape(nk,nb,-1).conj()  @ S_me.reshape(nk,nb,-1).transpose(0,2,1)
        ## Take care of time reversal 
        S_me[time_rev] = S_me[time_rev].conj()
        return S_me

    def get_iBZ_wf(self, ik):
        """
        Get wavefunctions and G-vectors for a specific k-point.

        Args:
            ik (int): iBZ K-point index.

        Returns:
            list: Wavefunctions at ik (nspin,nbands,nspinor,ngvec) and G-vectors (ngvec,3) 
            in crystal coordinates.
        """
        self.assert_k_inrange(ik)
        return [self.wf[ik][..., :self.ngvecs[ik]], self.gvecs[ik, :self.ngvecs[ik], :]]


    def wfcG2r(self, ik, ib, grid=[]):
        """
        Convert wavefunctions from G-space to real space.

        Args:
            ik (int): iBZ K-point index.
            ib (int): Band index.
            grid (list, optional): FFT grid size. Defaults to the default FFT box.

        Returns:
            numpy.ndarray: Wavefunctions in real space.
        """
        assert len(grid) == 0 or len(grid) == 3, "Grid must be an empty list or a list of 3 integers"
        self.assert_k_inrange(ik)
        self.assert_bnd_range(ib)

        if len(grid) == 0:
            grid = self.fft_box
        else:
            for i in range(3):
                assert grid[i] >= self.fft_box[i], f"Invalid FFT grid. Grid must be >= {self.fft_box[i]}"

        print(f'FFT Grid: {grid[0]} {grid[1]} {grid[2]}')

        wfc_tmp = self.wf[ik][:, ib, :, :self.ngvecs[ik]]
        gvec_tmp = self.gvecs[ik, :self.ngvecs[ik], :]
        return self.to_real_space(wfc_tmp, gvec_tmp, grid=grid)

    def write2cube(self, ik, ib, grid=[]):
        """
        Write wavefunctions to a cube file.

        Args:
            ik (int): iBZ K-point index.
            ib (int): Band index.
            grid (list, optional): FFT grid size. Defaults to the default FFT box.
        """
        wfc_r = self.wfcG2r(ik, ib, grid=grid)
        wfc_r = np.sum(np.abs(wfc_r) ** 2, axis=(1))
        wfc_r = wfc_r / wfc_r.max()
        for ispin in range(self.nspin):
            filename = f'wfc_k{ik + 1}_bnd_{ib + self.min_bnd + 1}_spin{ispin + 1}.cube'
            write_cube(filename, wfc_r[ispin], self.ydb.lat.T, self.ydb.car_atomic_positions,
                       self.ydb.atomic_numbers, header='Real space electronic wavefunction')

    def get_iBZ_kpt(self, ik):
        """
        Get the iBZ k-point in crystal coordinates.

        Args:
            ik (int): iBZ K-point index.

        Returns:
            numpy.ndarray: K-point in crystal coordinates.
        """
        self.assert_k_inrange(ik)
        return self.kpts_iBZ[ik]

    def rotate_wfc(self, ik, isym):
        """
        Rotate wavefunctions using a symmetry operation.

        Args:
            ik (int): iBZ K-point index.
            isym (int): Symmetry operation index from SAVE.

        Returns:
            list: Rotated wavefunctions and G-vectors.
        """
        wfc_k, gvecs_k = self.get_iBZ_wf(ik)
        sym_mat = self.ydb.sym_car[isym]
        kvec = self.get_iBZ_kpt(ik)
        time_rev = (isym >= len(self.ydb.sym_car) / (1 + int(np.rint(self.ydb.time_rev))))
        return self.apply_symm(kvec, wfc_k, gvecs_k, time_rev, sym_mat)


    @func_profile
    def apply_symm(self, kvec, wfc_k, gvecs_k, time_rev, sym_mat, frac_vec=np.array([0, 0, 0])):
        """
        Apply symmetry to wavefunctions.

        Args:
            kvec (numpy.ndarray): K-point in crystal coordinates.
            wfc_k (numpy.ndarray): Wavefunctions for the k-point.
            gvecs_k (numpy.ndarray): G-vectors for the k-point.
            time_rev (bool): Whether to apply time reversal.
            sym_mat (numpy.ndarray): Symmetry matrix. (Cart units)
            frac_vec (numpy.ndarray, optional): Fractional translation vector. Defaults to [0, 0, 0]. (cart units)

        Returns:
            list: Rotated wavefunctions and G-vectors.
        """
        sym_red = np.linalg.inv(self.ydb.lat.T) @ sym_mat.T @ self.ydb.lat.T
        sym_red = np.rint(sym_red).astype(int)
        gvec_rot = gvecs_k @ sym_red

        if self.nspinor == 2:
            su_mat = su2_mat(sym_mat, time_rev).astype(wfc_k.dtype)
            #'ij,sbjg->sbig', su_mat, wfc_k
            wfc_rot = su_mat[None,None,:,:]@wfc_k
        else : wfc_rot = wfc_k.copy()

        # Add phase due to fractional translation
        Rkvec = sym_red.T @ kvec
        tau_frac = np.linalg.inv(self.ydb.lat.T) @ frac_vec
        kphase = np.exp(-1j * 2 * np.pi * np.dot(Rkvec, tau_frac))
        gphase = kphase * np.exp(-1j * 2 * np.pi * (gvec_rot @ tau_frac))
        wfc_rot *= gphase[None, None, None, :]

        if time_rev:
            wfc_rot = wfc_rot.conj()

        return [Rkvec, wfc_rot, gvec_rot]


    @func_profile
    def to_real_space(self, wfc_tmp, gvec_tmp, grid=[]):
        """
        Convert wavefunctions from G-space to real space.

        Args:
            wfc_tmp (numpy.ndarray): Wavefunctions in G-space.
            gvec_tmp (numpy.ndarray): G-vectors.
            grid (list, optional): FFT grid size. Defaults to the default FFT box.

        Returns:
            numpy.ndarray: Wavefunctions in real space.
        """
        if len(grid) == 0:
            grid = self.fft_box
        cel_vol = abs(np.linalg.det(self.ydb.lat.T))

        tmp_wfc = np.zeros((len(wfc_tmp), self.nspinor, grid[0], grid[1], grid[2]), dtype=wfc_tmp.dtype)

        Nx_vals = np.where(gvec_tmp[:, 0] >= 0, gvec_tmp[:, 0], gvec_tmp[:, 0] + grid[0])
        Ny_vals = np.where(gvec_tmp[:, 1] >= 0, gvec_tmp[:, 1], gvec_tmp[:, 1] + grid[1])
        Nz_vals = np.where(gvec_tmp[:, 2] >= 0, gvec_tmp[:, 2], gvec_tmp[:, 2] + grid[2])
        #
        assert np.min(Nx_vals) >=0 and (np.min(Ny_vals) >=0 and np.min(Nz_vals)) >=0, "Wrong fft indices"
        tmp_wfc[:,:,Nx_vals, Ny_vals,Nz_vals] = wfc_tmp
        return scipy.fft.ifftn(tmp_wfc,norm="forward",axes=(2,3,4))/np.sqrt(cel_vol)

    
    def expand_fullBZ(self):
        """
        Expands the wavefunctions to the full Brillouin Zone (BZ) by applying symmetry operations.

        This method constructs wavefunctions at all symmetry-equivalent k-points in the full BZ.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        #
        if getattr(self, 'wf_bz', None) is not None: return
        #
        kpt_idx = self.ydb.kpoints_indexes
        sym_idx = self.ydb.symmetry_indexes
        nkBZ = len(sym_idx)

        self.wf_bz = np.zeros((nkBZ, self.nspin, self.nbands, self.nspinor, self.ng),dtype=self.wf.dtype)
        self.g_bz = 2147483646 + np.zeros((nkBZ,self.ng,3),dtype=int)
        self.kBZ = np.zeros((nkBZ,3))
        # NM : The reason we want to replace the existing kBZ variable is to make sure we have 
        # correct rotated kpoint (here they should not differ by a G vector !)
        self.ngBZ = np.zeros(nkBZ,dtype=int)
        for i in tqdm(range(nkBZ), desc="Expanding Wavefunctions full BZ"):
            ik = kpt_idx[i]
            isym = sym_idx[i]
            #if isym == 1 : w_t, g_t = self.get_iBZ_wf(ik)
            kbz, w_t, g_t = self.rotate_wfc(ik, isym)
            ng_t = w_t.shape[-1]
            self.ngBZ[i] = ng_t
            self.kBZ[i] = kbz
            self.wf_bz[i][...,:ng_t]  = w_t
            self.g_bz[i][:ng_t,:]  = g_t
        #
        self.ktree = build_ktree(self.kBZ)
        return 

    def get_BZ_kpt(self, ik):
        """
        Get the BZ k-point in crystal coordinates.

        Args:
            ik (int): BZ K-point index.

        Returns:
            numpy.ndarray: K-point in crystal coordinates.
        """
        return self.kBZ[ik]

    def get_BZ_wf(self, ik):
        """
        Get wavefunctions and G-vectors for a specific k-point in  full BZ.
        
        Args:
            ik (int): BZ K-point index.
        Returns:
            list: Wavefunctions at ik (nspin,nbands,nspinor,ngvec) and G-vectors in crystal coordinates (ngvec,3).
        """
        if getattr(self, 'wf_bz', None) is None: self.expand_fullBZ()
        #
        return [self.wf_bz[ik][..., :self.ngBZ[ik]], self.g_bz[ik, :self.ngBZ[ik], :]]


    @func_profile
    @citation("M. Nalabothula et al. arXiv:2511.21540 (2025)")
    def Dmat(self, symm_mat=None, frac_vec=None, time_rev=None):
        """
        Computes the symmetry-adapted matrix elements < Rk | U(R) | k >.
        Implements Eq. 5 of M Nalabothula et al. arXiv:2511.21540.

        Parameters
        ----------
        symm_mat : np.ndarray
            Array of symmetry rotation matrices (nsym, 3, 3).
        frac_vec : np.ndarray
            Fractional translation vectors associated with symmetries (nsym, 3).
        time_rev : bool
            Whether time-reversal symmetry is included.

        if symm_mat or frac_vec or time_rev is None, the use the symmetries from SAVE
        Returns
        -------
        np.ndarray
            The computed D matrix with shape (nsym, nk, nspin, nbands, nbands).
        """
        # Compute < Rk | U(R) | k> matrix elements 
        ## isym >= len(symm_mat)/(1+int(time_rev)) must be timereversal symmetries
        ## X -> Rx + tau, R matrices are given in symm_mat, tau are frac_vec, time_rev is bool
        ## (nsym, nk, nspin, Rk_bnd, k_bnd)
        expand_wf_present = True
        if getattr(self, 'wf_bz', None) is None: expand_wf_present = False
        ##
        ## Check if already computed for SAVE symetries 
        dmat_save = getattr(self, 'save_Dmat', None)
        #
        #
        is_save_symm = False
        if symm_mat is None or frac_vec is None or time_rev is None:
            is_save_symm = True
        #
        if is_save_symm: 
            ## if dmats are already computed, return those
            if dmat_save is not None : return dmat_save
            #
            symm_mat = self.ydb.sym_car
            frac_vec = np.zeros((len(symm_mat),3),dtype=symm_mat.dtype)
            time_rev = int(np.rint(self.ydb.time_rev))
        #
        ktree = self.ktree #build_ktree(self.kBZ)
        Dmat = []
        nsym = len(symm_mat)
        kpt_idx = self.ydb.kpoints_indexes
        sym_idx = self.ydb.symmetry_indexes
        #
        assert nsym == len(frac_vec), "The number for frac translation must be same as Rotation matrices"
        for ik in tqdm(range(self.nkBZ), desc="Dmat"):
            # IN case the wfc are already expanded, load them
            if expand_wf_present:
                wfc_k, gvec_k = self.get_BZ_wf(ik)
                kvec = self.get_BZ_kpt(ik)
            else:
                ## else rotate them
                ixk = kpt_idx[ik]
                isk = sym_idx[ik]
                kvec, wfc_k, gvec_k = self.rotate_wfc(ixk, isk)

            for isym in range(nsym):
                trev = (isym >= nsym/(1+int(time_rev)))
                #
                ## Compute U(R)\psi_k
                #
                Rk, wfc_Rk, gvec_Rk = self.apply_symm(kvec, wfc_k, gvec_k, trev, symm_mat[isym], frac_vec[isym])
                idx = find_kpt(ktree, Rk)
                #
                ## get Rk wfc 
                # in case stored
                if expand_wf_present:
                    w_rk, g_rk = self.get_BZ_wf(idx)
                    k_rk = self.get_BZ_kpt(idx)
                else :
                    iktmp = kpt_idx[idx]
                    istmp = sym_idx[idx]
                    k_rk, w_rk, g_rk = self.rotate_wfc(iktmp, istmp)
                Dmat.append(wfc_inner_product(k_rk, w_rk, g_rk, Rk, wfc_Rk, gvec_Rk))
        #
        Dmat = np.array(Dmat).reshape(self.nkBZ, nsym, self.nspin, self.nbands, self.nbands).transpose(1,0,2,3,4)
        #
        if is_save_symm: self.save_Dmat = Dmat
        #
        return Dmat

    def OverlapUkkp(self, kpt_bra, kpt_ket):
        """
        Compute the following matrix elements : < k_bra | e^{i(k_bra-k_ket).r} | k_ket>
        in other words, it computes overlap of periodic parts of k_bra an k_ket
        """
        kpt_bra = np.array(kpt_bra)
        kpt_ket = np.array(kpt_ket)

        ikpt_ket = find_kpt(self.ktree,kpt_ket)
        ikpt_bra = find_kpt(self.ktree,kpt_bra)
        #
        kpt_idx = self.ydb.kpoints_indexes
        sym_idx = self.ydb.symmetry_indexes
        #
        ibz_ket = kpt_idx[ikpt_ket]
        isym_ket = sym_idx[ikpt_ket]
        #
        ibz_bra = kpt_idx[ikpt_bra]
        isym_bra = sym_idx[ikpt_bra]
        #
        ## get the wfcs:
        k_rk_ket, w_rk_ket, g_rk_ket = self.rotate_wfc(ibz_ket, isym_ket)
        k_rk_bra, w_rk_bra, g_rk_bra = self.rotate_wfc(ibz_bra, isym_bra)

        G0_ket = kpt_ket-k_rk_ket
        G0_bra = kpt_bra-k_rk_bra
        #
        G0 = G0_ket-G0_bra
        return wfc_inner_product(G0, w_rk_bra, g_rk_bra, np.array([0,0,0]), w_rk_ket, g_rk_ket)

## end of class

##
@func_profile
def wfc_inner_product(k_bra, wfc_bra, gvec_bra, k_ket, wfc_ket, gvec_ket, ket_Gtree=None):
    """
    Computes the inner product between two wavefunctions in reciprocal space. <k_bra | k_ket>
    
    Parameters
    ----------
    k_bra : ndarray
        Crystal momentum of the bra wavefunction (3,) in reduced coordinates.
    wfc_bra : ndarray
        Wavefunction coefficients for the bra state with shape (nspin, nbnd, nspinor, ng).
    gvec_bra : ndarray
        Miller indices of the bra wavefunction (ng, 3) in reduced coordinates.
    k_ket : ndarray
        Crystal momentum of the ket wavefunction (3,) in reduced coordinates.
    wfc_ket : ndarray
        Wavefunction coefficients for the ket state with shape (nspin, nbnd, nspinor, ng).
    gvec_ket : ndarray
        Miller indices of the ket wavefunction (ng, 3) in reduced coordinates.
    ket_Gtree  : scipy.spatial._kdtree.KDTree (optional)
        Kdtree for gvec_ket. leave it or give None to internally build one
    #
    Returns
    -------
    ndarray
        Inner product matrix of shape (nspin, nbnd, nbnd). If the momenta mismatch
        is too large, returns a zero matrix.
    """
    #
    # Check consistency of wavefunction dimensions
    assert wfc_ket.shape[:3] == wfc_bra.shape[:3], "Inconsistant wfcs"
    #
    nspin, nbnd, nspinor = wfc_ket.shape[:3]
    kdiff = k_bra-k_ket
    G0 = np.rint(kdiff).astype(int)
    kdiff = kdiff-G0
    ## crystal momentum mismatch delta_{k,k'}
    if np.max(np.abs(kdiff)) > 1e-5:
        return np.zeros((nspin, nbnd, nbnd),dtype=wfc_ket.dtype)
    # Construct KDTree for nearest-neighbor search in G-vectors
    if ket_Gtree is None:
        ket_Gtree = KDTree(gvec_ket)
    gbra_shift = gvec_bra + G0[None,:]
    ## get the nearest indices and their distance
    dd, ii = ket_Gtree.query(gbra_shift, k=1)
    #
    wfc_bra_tmp = np.zeros(wfc_ket.shape,dtype=wfc_ket.dtype)
    # Get only the indices that are present
    bra_idx = ii[dd < 1e-6]
    #
    wfc_bra_tmp[:,:,:,bra_idx] = wfc_bra[...,dd<1e-6].conj()
    # return the dot product
    inprod = np.zeros((nspin, nbnd, nbnd),dtype=wfc_bra.dtype)
    for ispin in range(nspin):
        inprod[ispin] = wfc_bra_tmp[ispin].reshape(nbnd,-1)@wfc_ket[ispin].reshape(nbnd,-1).T
    #return np.einsum('sixg,sjxg->sij',wfc_bra_tmp,wfc_ket,optimize=True) #// einsum is very slow
    return inprod


def su2_mat(symm_mat,time_rev1):
    """
    Computes the SU(2) spinor rotation matrix for a given symmetry matrix.

    Parameters
    ----------
    symm_mat : ndarray
        3x3 rotation matrix describing the symmetry operation.
    time_rev1 : bool
        If True, applies time-reversal symmetry to the SU(2) matrix.

    Returns
    -------
    ndarray
        2x2 SU(2) matrix representing the spinor transformation.
    """
    # If the symmetry matrix corresponds to identity or inversion, return identity matrix
    if np.isclose(np.abs(np.trace(symm_mat)),3,atol=1e-03):
        su_mat = np.eye(2)
    else :
        sigma_0= np.array([1, 0 , 0, 1]).reshape(2,2)
        sigma_x= np.array([0, 1 , 1, 0]).reshape(2,2)
        sigma_y= np.array([0,-1j, 1j,0]).reshape(2,2)
        sigma_z= np.array([1 , 0, 0,-1]).reshape(2,2)
        sym_temp = symm_mat/np.linalg.det(symm_mat)

        if abs(sym_temp[2,0] +1) < 10**-5 :
            alpha = 0
            beta = -np.pi/2
            delta = np.arctan2(sym_temp[0,1],sym_temp[0,2])
        elif abs(sym_temp[2,0] - 1) < 10**-5:
            alpha = 0
            beta = np.pi/2
            delta = np.arctan2(-sym_temp[0,1],-sym_temp[0,2])
        else :
            beta = np.arcsin(sym_temp[2,0])
            alpha = np.arctan2(sym_temp[1,0]/np.cos(beta),sym_temp[0,0]/np.cos(beta))
            delta = np.arctan2(sym_temp[2,1]/np.cos(beta),sym_temp[2,2]/np.cos(beta))
        #
        # Convert angles to half-angles for spinor representation
        alpha = alpha/2 ; beta = -beta/2 ; delta = delta/2 ;
        #
        # Compute SU(2) rotation matrix using Euler angle decomposition
        spin_RX_delta=sigma_0*np.cos(delta)-1j*sigma_x*np.sin(delta)
        spin_RY_beta =sigma_0*np.cos(beta) -1j*sigma_y*np.sin(beta)
        spin_RZ_alpha=sigma_0*np.cos(alpha)-1j*sigma_z*np.sin(alpha)
        su_mat = spin_RZ_alpha@spin_RY_beta@spin_RX_delta
    # Apply time-reversal symmetry if needed
    if time_rev1:
        su_mat = np.array([[0, -1], [1, 0]])@su_mat
    return su_mat

## NM : This need's to be fixed But Why do we need this?
    #def write(self,path):
    #    """
    #    Write the (new?) wavefunctions in new files
    #    """
    #    if os.path.isdir(path): shutil.rmtree(path)
    #    os.mkdir(path)

    #    #copy all the files
    #    oldpath = self.path
    #    filename = self.filename
    #    shutil.copyfile("%s/%s"%(oldpath,filename),"%s/%s"%(path,filename))
    #    for nk in range(self.nkpoints):
    #        fname = "%s_fragments_%d_1"%(filename,nk+1)
    #        shutil.copyfile("%s/%s"%(oldpath,fname),"%s/%s"%(path,fname))

    #    #edit with the new wfs
    #    wf = self.wf
    #    for nk in range(self.nkpoints):
    #        fname = "%s_fragments_%d_1"%(filename,nk+1)
    #        database = Dataset("%s/%s"%(path,fname),'r+')
    #        aux = np.array([wf[nk].real,wf[nk].imag])
    #        aux = np.moveaxis(aux,0,-1)
    #        database.variables['WF_COMPONENTS_@_SP_POL1_K%d_BAND_GRP_1'%(nk+1)][:] = aux
    #        database.close()
    #    print('New wavefunctions written in %s'%path)




if __name__ == "__main__":
    ywf = YamboWFDB(path='database')






##
# Authors: MN FP
##

import numpy as np
from yambopy.units import ha2ev
from yambopy.tools.funcs import bose,boltzman_f
from tqdm import tqdm
from yambopy.tools.citations import citation
#from numba import prange#,njit

#@njit(cache=True, nogil=True, parallel=True)
@citation("M. Zanfrognini et al. Phys. Rev. Lett. 131, 206902 (2023)")
def exc_ph_luminescence(ph_temp,ph_energies,exc_energies,exc_dipoles,exc_ph_mat_el,\
                        exc_energies_in=None,exc_temp=None,ph_channels='b',\
                        PL_energy_prefactor='PT',nexc_out='all',nexc_in='all',\
                        emin=0,emax=10,estep=0.01,broad=0.1,broad_0=None):
    """
    This function calculates the phonon-assisted satellites (phonon replica) in the 
    luminescence of indirect materials (i.e., phonon-mediated exciton recombination process).

    It uses a perturbation theory derivation that includes the off-diagonal elements
    of the exciton-phonon matrix elements.

    Exciton populations are approximated with a Boltzmann distribution function.

    See:
    - Phys. Rev. Materials 7, 024006 (2023)
    - Phys. Rev. Letters 131, 206902 (2023) Equations from (1) to (4)

    Returns:
    ----------
    * Energy axis in specified range (in eV)
    * Luminescence intensity

    Parameters
    ----------
    ph_temp : float
        Lattice temperature in kelvin
    ph_energies : float ndarray
        Phonon energies in eV [nqpts,nmodes]
    exc_energies : float ndarray
        Exciton energies in eV [nqpts,nexc_out] 
        Typically, YamboExcitonDB.eigenvalues are used
    exc_dipoles : cmplx ndarray
        Exciton dipole matrix elements in a.u. (bohr) 
        * If shape is [nexc_in], assume projected dipoles: typically, YamboExcitonDB.r_residuals are used.
        * If shape is [pol,nexc_in], perform polarization average. This can be produced with exc_dip_pol function in yambopy.bse.excitondipoles
    exc_ph_mat_el : cmplx ndarray
        Exciton-phonon coupling matrix elements in a.u. (hartree) [nqpts,nmodes,nexc_in,nexc_out]
        Typically, computed via exciton_phonon_matelem
    exc_energies_in : float ndarray, optional
        If given, zero-momentum states will be taken from this array. Default is exc_energies.
    exc_temp : float
        If given, effective temperature controlling excitonic population. Default is ph_temp.
    ph_channels : str
        'e'  -> only phonon emission channel
        'a' -> only phonon absorption channel
        'b' -> both channels (default)
    PL_energy_prefactor : str
        'PT' -> prefactor from time-dependent perturbation theory (default)
        'RS' -> prefactor from van Roosbroeck-Shockley relation (not including refraction index).
        None or any other str -> no prefactor is used
    nexc_out : int
        Number of finite-q excitons in the sum. Default: all available states.
    nexc_in : int
        Number of optical excitons in the sum. Default: all available states.
    emin, emax, esteps : float
        Laser energy range parameters in eV.
    broad : float
        Broadening parameter in eV.
    broad_0 : float
        Broadening parameter used inside satellite oscillator strengths in eV. Default is broad.
    """
    def get_PL_satellite(W,ph_sign=-1):
        """ Actual calculation of PL satellites at each frequency.
            Uses the same variables as main function.

            W --> laser energy value in hartree
            ph_sign -->
                * -1 : phonon emission
                * +1 : phonon absorption
        """
        # Satellite oscillator strenghts
        ## Dimensions: [nqpts, nmodes, nexc_in, nexc_out]
        satellite_energy = exc_energies[:,None,None,:]-exc_energies_in[None,None,:,None] \
                            + ph_sign*ph_energies[:,:,None,None] + 1j*broad_0
        satellite_energy_Ha = satellite_energy/ha2ev
        satellite_weight = exc_ph_mat_el/satellite_energy_Ha
    
        # Phonon emission/absorption factor: if em --> ph_occ+1, if abs --> ph_occ
        ph_fac = ph_occ - (ph_sign-1)/2.
    
        # Sum over exc_in and then square (this includes off-diagonal excph SE terms)
        T = np.einsum('ri,qmio->rqmo',exc_dipoles,satellite_weight,optimize=True)
        T = np.sum( np.abs(T)**2, axis=0)/3. # Average over pol. directions as well
        # Include occupation factors
        T = T * exc_occ[:,None,:]*ph_fac[:,:,None]

        # Energy differences [nqpts,nmodes,nexc_out]
        pole_energy = exc_energies[:,None,:]+ph_sign*ph_energies[:,:,None]
        pole_energy_Ha = pole_energy[:,:,:]/ha2ev
        
        # PL energy prefactors
        # Note: with 'PT+RS' as argument both will be considered
        light_fac = 1.
        if 'PT' in PL_energy_prefactor: # [nqpts,nmodes,nexc_out]
            light_fac = light_fac * 1./pole_energy_Ha
        if 'RS' in PL_energy_prefactor: # [nqpts,nmodes,nexc_out]
            light_fac = light_fac * W * (W-ph_sign*ph_energies[:,:,None])**2.

        # Energy conservation
        delta_funct = 1./((W-pole_energy_Ha)**2.+ broad_Ha**2.)
        E = light_fac*delta_funct
        # Dimensions
        E = E * broad_Ha/np.pi/Nqpts * ha2ev**2

        # Putting everything together
        return np.einsum('qmo,qmo->', E, T, optimize=True) 

    # Checks
    assert exc_energies.shape[0]==ph_energies.shape[0], "Q-point mismatch between excitons and phonons"
    Nqpts = ph_energies.shape[0]
    if exc_energies_in is None: exc_energies_in = exc_energies[0]
    if np.iscomplexobj(exc_energies):    exc_energies    = exc_energies.real
    if np.iscomplexobj(exc_energies_in): exc_energies_in = exc_energies_in.real
    assert ph_channels in ['b', 'e', 'a'], "Allowed values for phonon channels are 'b', 'e', 'a'"
    nexc_out_avail = min(exc_energies.shape[1],exc_ph_mat_el.shape[3])
    nexc_in_avail  = min(len(exc_energies_in),exc_dipoles.shape[-1],exc_ph_mat_el.shape[2])
    assert nexc_out <= nexc_out_avail, "less exciton states than requested (nexc_out)"
    assert nexc_in  <= nexc_in_avail,  "less exciton states than requested (nexc_in)"
    if nexc_out =='all': nexc_out = nexc_out_avail
    if nexc_in  =='all': nexc_in  = nexc_in_avail
    exc_energies    = exc_energies[:,:nexc_out]
    exc_energies_in = exc_energies_in[:nexc_in]
    exc_dipoles     = exc_dipoles[...,:nexc_in]
    # We conj because we need it for photon emission
    exc_ph_mat_el = np.conj( exc_ph_mat_el[...,:nexc_in,:nexc_out] ) 
    assert ph_energies.shape[1]==exc_ph_mat_el.shape[1], "number of modes mismatch between phonon energies and matrix elements"
    if broad_0 is None: broad_0 = broad # This goes in |E1-E2+i*broad_0|^(-2)
    broad = broad/2 # We are using explicit Lorentzian shape
    if exc_temp is None: exc_temp = ph_temp
    # We expect the dipoles to be unpolarized (3,nexc_in). If not, we expect projected dipoles
    # We extend the projected dipoles for compatibility
    if exc_dipoles.ndim==1: exc_dipoles = np.broadcast_to(exc_dipoles, (3, nexc_in)) 
    # Acoustic phonons at q=0: they are already excluded as exc_ph_mat_el[0,0:3,...]=czero

    # Occupation functions
    exc_min_energy = np.min(exc_energies)
    exc_occ = boltzman_f(exc_energies-exc_min_energy,exc_temp)
    ph_occ  = bose(ph_energies,ph_temp)

    # Laser energies
    light_energies = np.arange(emin,emax,estep,dtype=np.float32)
    nfreqs = len(light_energies)
    light_energies_Ha = light_energies/ha2ev#light_energies[None,None,None,:]/ha2ev
    broad_Ha = broad/ha2ev

    # Calculation
    PL_satellites = np.zeros(nfreqs)
    for w in tqdm(range(nfreqs),desc="Luminescence"):
        # Accumulate phonon emission satellites
        if ph_channels=='e' or ph_channels=='b': 
            PL_satellites[w] += get_PL_satellite(light_energies_Ha[w],ph_sign=-1)
        # Accumulate phonon absorption satellites
        if ph_channels=='a' or ph_channels=='b': 
            PL_satellites[w] += get_PL_satellite(light_energies_Ha[w],ph_sign=+1)

    return light_energies, PL_satellites



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
"""
submodule with classes to read BSE optical absorption spectra claculations
and information about the excitonic states
"""



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: MZ, FP
# First version by MZ (2022), revised and expanded by FP (2023)
#
# This file is part of the yambopy project
#
import numpy as np
import os
import shutil
from netCDF4 import Dataset
from yambopy.lattice import car_red, vec_in_list
from yambopy.kpoints import expand_kpoints
from yambopy.tools.string import marquee

def find_inversion_type(n_atoms,atom_pos,syms):
    """
    Check if the crystal has spatial inversion symmetry.

    Input:
    :: n_atoms is an array with the number of atoms per species
    :: atom_pos is an array of atomic positions per atom per species, already written in the COM frame

    Output:
    :: inv_type : 'spatial' or 'trev'
    :: inv_index: index of -1 symmetry
    """

    # Search for spatial inversion
    for i_species in range(len(n_atoms)): # N. species
        for i_atom in range(n_atoms[i_species]): # N. atoms/species
            atom1 = atom_pos[i_species][i_atom]
            for j_atom in range(n_atoms[i_species]):
                inv_type='spatial' # Assume we have it
                atom2 = atom_pos[i_species][j_atom]
                if np.allclose(atom1,-atom2): break            
                inv_type='trev' # If we get here, we don't have it

    # Search for inversion symmetry index
    for i_s,sym in enumerate(syms):
        if np.allclose(sym, -1.*np.identity(3)):
            inv_index = i_s
            break
    
    return inv_type, inv_index

class YamboEm1sRotate():
    """
    This class expands the em1s computed by yambo in the IBZ to the
    full BZ (i.e., Wigner-Seitz cell). 

    === Usage and variables ===

    >> yem1s   = YamboStaticScreeningDB(save=db_path,em1s=db_path)
    >> yexpand = YamboEm1sRotate(yem1s,save_path=db_path,path_output_DBs=db_path.split('/')[0]+'/Expanded',verbose=1)

    :: Input 
    - yem1s           -> YamboStaticScreeningDB object (IBZ calculation)
    - save_path       -> [OPTIONAL] Location of ns.db1 database (default is inside SAVE)
    - path_output_DBs -> [OPTIONAL] Print expanded databases at location (default: do not print)
    - verbose         -> [OPTIONAL] If True, prints a list of kpoints for QE calculation of nosym system

    :: Output
    - numpy array with expanded static screening 
    - [OPTIONAL] ndb.em1s and ndb.em1s_fragment_* databases corresponding to the 
    full BZ.
    - [OPTIONAL] data file 'kpoints_bz.dat' containing expanded rlu kpoint coordinates in PW format

    """

    def __init__(self,yem1s,save_path="SAVE",db1='ns.db1',path_output_DBs=None,verbose=0):

        # Attributes imported from StaticScreeningDB    
        supported_cutoffs = ['none','slab z']
        self.cutoff       = yem1s.cutoff

        if yem1s.filename != 'ndb.em1s': raise NotImplementedError("[ERROR] The screening rotation is only implemented for ndb.em1s.")
        if yem1s.cutoff not in supported_cutoffs: raise NotImplementedError("[ERROR] The em1s rotation is not currently implemented for cutoff %s."%yem1s.cutoff)

        self.rlat         = yem1s.rlat
        self.alat         = yem1s.alat
        self.qpoints_ibz  = yem1s.car_qpoints
        self.nqpoints_ibz = yem1s.nqpoints
        self.gvectors     = yem1s.gvectors
        self.red_gvectors = yem1s.red_gvectors
        self.ngvectors    = yem1s.ngvectors
        self.X_ibz        = yem1s.X
        self.em1s_path    = yem1s.em1s

        self.k_output     = 'kpoints_bz.dat'

        # Get symmetries in CC and real-space atomic positions
        if not os.path.isfile('%s/%s'%(save_path,db1)): raise FileNotFoundError("File %s not found."%db1)
        database = Dataset("%s/%s"%(save_path,db1), 'r')
        self.sym_car = np.transpose( database.variables['SYMMETRY'][:], (0,2,1) ) # transpose leaving first axis as symm index
        n_atoms =  database.variables['N_ATOMS'][:].astype(int)
        atom_pos = database.variables['ATOM_POS'][:]
        if verbose: iku_kpoints_ibz = database.variables['K-POINTS'][:].T
        database.close()
        self.nsyms = len(self.sym_car)

        print("=== Rotating em1s... ===")
        print(" * Getting q-map...  ")

        # Obtain transformed qpoints q'=Sq in the full BZ
        self.qpoints, self.qpoints_indices, self.syms_indices, _ = self.expand_kpoints(self.qpoints_ibz,self.sym_car,self.rlat)
        self.nqpoints = len(self.qpoints)

        print(" * Getting G-map ...  ")
        # Obtain transformed gvectors G'=S^{-1}G
        self.Sm1G_table = self.inverse_Gvector_table()

        print(" * Getting new em1s... ")

        # Spatial inversion or T-rev?
        # [WARNING] We assume one of the two is used!
        self.inv_type, self.inv_index = \
        find_inversion_type(n_atoms,atom_pos,self.sym_car)

        # Rotate em1s from IBZ to BZ
        self.rotate_em1s()

        if path_output_DBs is not None:
            print(" * Saving databases... [Remember that you might have to change the serial number!]")
            self.outpath = path_output_DBs
            self.saveDBS()
        else: print(" [!] Enter value for path_output_DBs to print expanded ndb.em1s")

        # Print kpts for PW nosym run
        if verbose: self.print_kpts_PW_format(iku_kpoints_ibz,units='rlu')

        print("===      Done.       ===")

    def expand_kpoints(self,kpoints,syms,rlat,atol=1e-6,verbose=0):
        """
        Wrapper of expand_kpoints from yambopy/kpoints.py

        Does not require any self. variables to be set.
        """

        weights, kpoints_indexes, symmetry_indexes, kpoints_full = expand_kpoints(kpoints,syms,rlat,atol=atol)

        if verbose: print("%d kpoints expanded to %d"%(len(kpoints),len(kpoints_full)))

        #set the variables
        expanded_car_kpoints  = kpoints_full
        kpoints_indices       = kpoints_indexes
        symmetry_indices      = symmetry_indexes
        weights_ibz           = weights

        return expanded_car_kpoints,kpoints_indices,symmetry_indices,weights_ibz

    def inverse_Gvector_table(self,tol=1e-5):
        """
        Build table Sm1G_table such as:

        if ig_S = Sm1G_table[ig,iS], then S^{-1}G[ig]=G[ig_S]

        - kwargs are atol and rtol for np.isclose
        """
        inv_syms = np.linalg.inv(self.sym_car)
        Sm1G_table = np.zeros((self.ngvectors,len(inv_syms)),dtype=int)

        self.rotated_gvectors = np.zeros([len(self.sym_car),len(self.gvectors),3])
        for iG,G in enumerate(self.gvectors):
            for i_S,sym in enumerate(inv_syms):
                check = np.sum( (self.gvectors - np.dot(sym,G))**2., axis=1) < tol
                if np.sum(check) == 1.0: # One G-vector G' has been found to correspond to sym^{-1}G
                    Sm1G_table[iG,i_S]=int(np.where(check==1.0)[0])
                elif np.sum(check)!= 1.0: #None or multiple G-vectors have been found
                    raise ValueError("\n[ERROR] Problem in mapping inverse G-vectors. Try:\n - (i) reducing isclose() tolerance in get_g_index (easy case) \n - (ii) check that yambo packs G-shells correctly for your lattice type and G-cutoff (difficult case)")

        return Sm1G_table

    def rotate_em1s(self):
        """
        Rotation of static screening

        The quantity rotated (content of ndb.em1s) is
 
        :math: D_{g1,g2}(q) = √v_g1(q) X_{g1g2}(q) √v_g2(q)   

        The rotations are performed according to

        - no time reversal (S)

            :math: D_{g1,g2}(Sq) = D_{S^-1g1,S^-1g2}(q)

        - time reversal is used (S->IS)

            :math:  D_{g1,g2}(ISq) = [ D_{(IS)^-1g1,(IS)^-1g2}(q) ]^*

        """
        X = np.zeros([self.nqpoints,self.ngvectors,self.ngvectors],dtype=np.complex64)
        
        for iq in range(self.nqpoints):
            iq_ibz = self.qpoints_indices[iq] # Index of untransformed q_ibz
            iS     = self.syms_indices[iq]    # Index of symmetry Sq=q_ibz
            for ig1 in range(self.ngvectors):
                Sm1_ig1 = self.Sm1G_table[ig1,iS] # index of G' such as G'=S-1G    
                for ig2 in range(self.ngvectors):
                    Sm1_ig2 = self.Sm1G_table[ig2,iS] 
                    # No TR
                    if self.inv_type=='spatial' or iS<self.inv_index:
                        X[iq,ig1,ig2]=self.X_ibz[iq_ibz,Sm1_ig1,Sm1_ig2]
                    # TR
                    if self.inv_type=='trev' and iS>=self.inv_index:
                        X[iq,ig1,ig2]=np.conj(self.X_ibz[iq_ibz,Sm1_ig1,Sm1_ig2])

        self.X = X                   

    def saveDBS(self):
        """
        Write yambo-compatible ndb.em1s and ndb.em1s_fragment* 
        databases containing the expanded static screening.

        TODO: 
        - enable restarts
        - automatically change serial number to new SAVE

        WARNING:
            There are many quirks and special rules and version-dependent ifs in the yambo IO of ndb.em1s 
            and its fragments.

            The implementation here has to overcome the difficulty of indulging these quirks so that
            the yambopy-created database is not rejected by yambo.
            
            The present implementation tries to be as general as reasonably possible, and it has been tested,
            but instances may still arise with the appearance of errors and incompatibilities while trying to 
            get yambo to read these dbs. Also, in the event of an update in the yambo IO of ndb.em1s, these 
            writing functions will almost certainly be broken. 
        """
        path = self.outpath

        # If it exists, always remove outdir for safety
        if os.path.isdir(path): shutil.rmtree(path)
        os.mkdir(path)

        self.write_em1s_header(path)

        self.write_em1s_fragments(path)

    def write_em1s_header(self,path):
        """ Write ndb.em1s

            :: Dimensions ([F]ixed or [V]ariable):
               0 - D_0000000003 = 3   [F]-> HEAD_VERSION, HEAD_QPT, X_RL_vecs
               1 - D_0000000001 = 1   [F]-> HEAD_REVISION, SERIAL_NUMBER, HEAD_WF, CUTOFF, FRAGMENTED,
                                            Xs_energies_xc_KIND, Xs_wavefunctions_xc_KIND, GAUGE, 
                                            X_Time_ordering, X_TDDFT_KERNEL, X_OPTICAL_AVERAGE
               2 - D_0000000002 = 2   [F]-> SPIN_VARS, TEMPERATURES, X_DRUDE
               3 - D_0000000004 = 4   [F]-> HEAD_R_LATT
               4 - D_00000000Nk = Nk  [V]-> HEAD_QPT
               5 - D_0000000100 = 100 [F]-> CUTOFF, Xs_energies_xc_KIND, Xs_wavefunctions_xc_KIND,
                                            GAUGE, X_Time_ordering, X_TDDFT_KERNEL, X_OPTICAL_AVERAGE
               6 - D_0000000005 = 5   [F]-> X_PARS_1
               7 - D_0000000010 = 10  [F]-> X_PARS_3
               8 - D_00000000NG = NG  [V]-> X_RL_vecs

            :: Contents of X_PARS_1:
                - [V]X matrix size, X%ng 
                - [F]X band range, tmp_ib (start, end)
                - [F]X e/h energy range, X%ehe (start, end)  

            :: Contents of X_PARS_2: [Not usually written, we assume it's not present]

            :: Contents of X_PARS_3:
                - [F]X poles, X%cg_percentual,
                - [V]RL vectors in the sum, X%ngostnts
                - [F][r,Vnl] included, X%Vnl_included
                - [F]Longitudinal Gauge, local_long_gauge [normally absent]
                - [F]Field direction, X%q0 (x,y,z)
                - [F]BZ energy Double Grid, use_X_DbGd
                - [F]BZ energy DbGd points, X_DbGd_nkpts
                - [F]BZ Q point size factor, X_DbGd_percentual

        """
        def netcdftype(var_type):
            """ Distinguish between double and float
            """
            if var_type=='float32': return 'f4'
            elif var_type=='float64': return 'f8'
            else: raise TypeError('\n[ERROR] Variable type not recognized. It should be either float (float32) or double (float64).\n')
        
        # New database
        dbs = Dataset(path+'/ndb.em1s',mode='w',format='NETCDF4')

        # Old database
        dbs_ibz = Dataset(self.em1s_path+'/ndb.em1s')
        ibz_dims = dbs_ibz.dimensions.values()
        ibz_vars = dbs_ibz.variables.values() 

        # New dimensions

        # check if we are in an unfortunate cases due to how yambo writes the databases
        hard_dims = [3,1,2,4,100,5,10]  # fixed dimension sizes
        ibz_ = self.nqpoints_ibz in hard_dims #q_ibz number coincides with one of the fixed dimensions
        bz_  = self.nqpoints     in hard_dims #q_bz number coincides with one of the fixed dimensions
        gv_  = self.ngvectors    in hard_dims #g-vecs number coincides with one of the fixed dimensions

        sizes     = [dim.size for dim in ibz_dims]
        n_dims    = len(sizes)        
        ind_q_ibz_dim = sizes.index(self.nqpoints_ibz) 
        # Nq-related dimension is in fifth place, Gv-related one in last place       
        error_conditions= np.array([n_dims==9 and ind_q_ibz_dim!=4, n_dims<9 and ibz_==False and gv_==False]) 
        if error_conditions.any(): raise ValueError("[ERROR] something wrong with em1s dbs dimensions.")
        
        # Create dimensions
        iaux=0
        for dim in ibz_dims:
            # manage new q_bz dimension
            if iaux==4: 
                if ibz_==False and bz_==False:
                    # Replace old Nq_ibz dimension with new Nq_bz one
                    dbs.createDimension('D_%010d'%self.nqpoints,self.nqpoints)
                    iaux+=1
                if ibz_==True and bz_==False: 
                    # Add new Nq_bz dimension in between existing ones
                    dbs.createDimension('D_%010d'%self.nqpoints,self.nqpoints) # Create Nq_bz
                    dbs.createDimension(dim.name,dim.size)
                    iaux+=1
                if ibz_==False and bz_==True: 
                    # Do not copy the old unneeded Nq_ibz dimension
                    iaux+=1
                if ibz_==True and bz_==True:
                    # Copy existing dimensions as normal
                    dbs.createDimension(dim.name,dim.size)
                    iaux+=1
            # just copy existing dimensions including G-size
            else:
                dbs.createDimension(dim.name,dim.size)
                iaux+=1

        # Create variables  
        for var in ibz_vars:
            if var.name=='HEAD_QPT': dbs.createVariable('HEAD_QPT',netcdftype(var.dtype), ('D_%.10d'%3, 'D_%.10d'%self.nqpoints))
            elif var.name=='HEAD_R_LATT': dbs.createVariable('HEAD_R_LATT',netcdftype(var.dtype), ('D_%.10d'%4))
            else: dbs.createVariable(var.name, var.dtype, var.dimensions)
            
        # Store values in new DB, including new qpt coords in iku
        for var in ibz_vars: 
            if var.name=='HEAD_QPT':      dbs[var.name][:] = np.array([q*self.alat for q in self.qpoints]).T
            elif var.name=='HEAD_R_LATT': dbs[var.name][:] = [ self.nqpoints for i in range(4) ]
            else:                         dbs[var.name][:] = dbs_ibz[var.name][:]

        dbs.close()
        dbs_ibz.close()

    def write_em1s_fragments(self,path):
        """ Write ndb.em1s_fragment*

            :: Dimensions ([F]ixed or [V]ariable):
               0 - D_0000000006 = 6   [F]-> FREQ_PARS_sec_iqN
               1 - D_0000000001 = 2   [F]-> FREQ_sec_iqN, X_Q_N
               2 - D_0000000002 = 1   [F]-> FREQ_sec_iqN, X_Q_N
               8 - D_00000000NG = NG  [V]-> X_Q_N

            :: Contents of FREQ_PARS_sec_iqN:
                - [V]Current Q-pt index, iq
                - [F]X energy range, Xw%er (start, end)
                - [F]X damping range, Xw%dr (start, end)
                - [F]Number of frequencies, Xw%n_freqs
        """

        # Old database (reference for q=1)
        dbs_ibz = Dataset(self.em1s_path+'/ndb.em1s_fragment_1')
        ibz_dims = dbs_ibz.dimensions.values()
        ibz_vars = dbs_ibz.variables.values()
        pars = dbs_ibz.variables['FREQ_PARS_sec_iq1'][:]
        freq = dbs_ibz.variables['FREQ_sec_iq1'][:]

        # Create each fragment
        for iq_bz in range(self.nqpoints):

            iq_aux = iq_bz+1
            pars[0] = iq_aux 

            dbs_qbz = Dataset(path+'/ndb.em1s_fragment_%d'%iq_aux,mode='w',format='NETCDF4') 

            # Dimensions
            for dim in ibz_dims: dbs_qbz.createDimension(dim.name,dim.size)

            # Variables
            for var in ibz_vars: 
                if 'PARS' in var.name: dbs_qbz.createVariable('FREQ_PARS_sec_iq%d'%iq_aux, var.dtype, var.dimensions)
                if 'Q_se' in var.name: dbs_qbz.createVariable('FREQ_sec_iq%d'%iq_aux, var.dtype, var.dimensions)
                if 'X' in var.name:    dbs_qbz.createVariable('X_Q_%d'%iq_aux, var.dtype, var.dimensions)

            # Values
            dbs_qbz['FREQ_PARS_sec_iq%d'%iq_aux][:] = pars   
            dbs_qbz['FREQ_sec_iq%d'%iq_aux][:]      = freq

            # Finally, store expanded screening
            re, im = [self.X[iq_bz].real, self.X[iq_bz].imag]
            dbs_qbz['X_Q_%d'%iq_aux][:] = np.stack((re,im),axis=2)

            dbs_qbz.close()

        dbs_ibz.close()

    def print_kpts_PW_format(self,kpoints_iku,units='rlu'):
        """
        Print expanded kpoints in the Wigner-Seitz cell for PW
        nosym=.true., noinv=.true. calculation

        - kpoints_iku: iku kpoints in the IBZ read from ns.db1
        - units: either rlu or cc
        """
        def format_array(value):
            if value >= 0: return ' '+format(value,'.9f')
            else:          return format(value,'.9f')

        print(" * Printing PW-format kpoints file.")

        kpoints_ibz = np.array([ k/self.alat for k in kpoints_iku ])
        kpoints     = self.expand_kpoints(kpoints_ibz,self.sym_car,self.rlat)[0]
        if units=='rlu': points = car_red(kpoints,self.rlat)
        if units=='cc':  points = kpoints*self.alat[0]

        kpts2prnt = np.empty((len(points),4),dtype=object)
        for i in range(len(points)):
            for j in range(4):
                if j<3: 
                    val = points[i,j]
                    if val>=0: kpts2prnt[i,j]=' '+"{:0.9f}".format(val)
                    else:      kpts2prnt[i,j]="{:0.9f}".format(val)   
                if j==3:       kpts2prnt[i,j]='1'

        np.savetxt(self.k_output, kpts2prnt, fmt='%s %s %s %s')

    def __str__(self):

        lines = []; app=lines.append
        app(marquee(self.__class__.__name__))

        app('nqpoints (ibz):   %d'%self.nqpoints_ibz)
        app('nqpoints (bz):   %d'%self.nqpoints)
        app('X size (G-space): %d'%self.ngvectors)
        app('cutoff:           %s'%self.cutoff)
        app('inversion type:   %s'%self.inv_type)
        app('nsymmetries:      %s'%self.nsyms)
        app('inversion sym is: %s'%self.inv_index)

        return "\n".join(lines)




# Copyright (c) 2018, Henrique Pereira Coutada Miranda
#
# All rights reserved.
#
# This file is part of the yambopy project
#
"""
This file contains a class to analyse the BSE absorption spectra

The features to be provided include: 
    1. Read the BSE diago file and plot chi including excitonic effects
    2. Obtain the excitonic states energy, degeneracies and optical intensity
    3. Create flows to run ypp and obtain the excitonic wavefunctions
    4. Produce a json file that can be used in the excitonwebsite

Long term:
    1. For the same exciton store multiple WFs changing the hole position

To initialize this structure we require instances of:
    - YamboExcitonDB (which already requires YamboLatticeDB)

The json format produced by this class for the excitonwebsite is:

structure stuff
    -lattice
    -atoms
    -atypes
bse stuff
    -eps - Absorption spectra
    -eel - Electron energy loss spectra
    -nx,ny,nz - the dimensions of the datagrid
    -datagrid
    -excitons - list of excitons to show

each exciton entry contains:
    -datagrid
    -intensity
    -index
    -energy

"""
from __future__ import print_function, division
import os
from yambopy.tools.string import marquee
from yambopy import *

class YamboBSEAbsorptionSpectra():
    """
    Create a file with information about the excitons from Yambo files
    """
    def __init__(self,excitondb):
        """
        Parameters:
            excitondb - an instance of the excitonDB class
        """
        self.excitondb = excitondb

    @classmethod
    def from_folder(cls,folder):
        raise NotImplementedError('from_folder not implemented yet')
        #try to find a excitonDB class
        #initialize this class
        return cls

    def get_excitons(self,min_intensity=0.1,max_energy=4,eps=1e-4):
        """ 
        Obtain the excitons using ypp
        Parameters:
            min_intensity - Only plot excitons with intensity larger than this value (default: 0.1)
            max_energy    - Only plot excitons with energy below this value (default: 4 eV)
            Degen_Step    - Only plot excitons whose energy is different by more that this value (default: 0.0)
        """
        excitons = self.excitondb.get_nondegenerate(eps=eps)

        #filter with energy
        excitons = excitons[excitons[:,0]<max_energy]

        #filter with intensity
        excitons = excitons[excitons[:,1]>min_intensity]

        #filter with degen
        if Degen_Step:

            #create a list with differences in energy
            new_excitons = []
            prev_exc = 0
            for exc in self.excitons:
                e,i,index = exc
                #if the energy of this exciton is too diferent then we add it to the list
                if abs(e-prev_exc)<Degen_Step:
                    new_excitons[-1][1] += i
                    continue
                new_excitons.append([e,i,index])
                intensity = 0
                prev_exc = e
            self.excitons = np.array(new_excitons)

        #create dictionary with excitons 
        excitons = self.data["excitons"]
        for e,intensity,i in self.excitons:
            exciton = {"energy": e,
                       "intensity": intensity,
                       "index": i}
            excitons.append(exciton)
        return self.excitons

    def get_wavefunctions(self, FFTGvecs=30,
                          Cells=[1,1,1], Hole=[0,0,0],
                          Direction="123", Format="x",
                          Degen_Step=0.0100,
                          MinWeight=1e-8,
                          repx=list(range(-1,2)), repy=list(range(-1,2)), repz=list(range(-1,2)),
                          wf=False):
        """
        Collect all the wavefuncitons with an intensity larger than self.threshold
        Parameters:
          FFTGvecs   - Number of FFTGvecs. Related to how accurate the representation is
          Cells      - Number of cells to plot in real space
          Hole       - Define the hole position in cartesian coordinates
          Direction  - Choose the directions to plot along
          Format     - Choose the format to plot in. Can be: x for xcrysden or g for gnuplot (default: 'x' for xcrysden)
          Degen_Step - Threshold to merge degenerate states. If the difference in energy between the states is smaller than
                       this value their wavefunctions will be plotted together
          repx       - Number or repetitions along the x direction
          repy       - Number or repetitions along the x direction
          repz       - Number or repetitions along the x direction
          wf         - Get the wavefuncitons in real space or not (default: False)
        """
        if self.excitons is None:
            raise ValueError( "Excitons not present. Run YamboBSEAbsorptionSpectra.get_excitons() first" )
        self.data["excitons"] = []

        #create a ypp file using YamboIn for reading the wavefunction
        yppwf = YamboIn('ypp -e w -V all',filename='ypp.in',folder=self.path)
        yppwf['Format'] = Format
        yppwf['Direction'] = Direction
        yppwf['FFTGvecs'] = [FFTGvecs,'Ry']
        yppwf['Degen_Step'] = [Degen_Step,'eV']
        yppwf['Hole'] = [Hole,'']
        yppwf['Cells'] = [Cells,'']

        #create a ypp file using YamboIn for reading the excitonic weights
        yppew = YamboIn('ypp -e a',filename='ypp.in',folder=self.path)
        yppew['MinWeight'] = MinWeight
        yppew['Degen_Step'] = Degen_Step

        keywords = ["lattice", "atoms", "atypes", "nx", "ny", "nz"]
        for exciton in self.excitons:
            #get info
            e,intensity,i = exciton

            if wf:
                ##############################################################
                # Excitonic Wavefunction
                ##############################################################
                #create ypp input for the wavefunction file and run
                yppwf["States"] = "%d - %d"%(i,i)
                yppwf.write("%s/yppwf_%d.in"%(self.path,i))

                filename = "o-%s.exc_%dd_%d%s"%(self.job_string,len(Direction),i,{"g":"","x":".xsf"}[Format] )
                print(filename)
                if not os.path.isfile(filename):
                    os.system("cd %s; ypp -F yppwf_%d.in -J %s"%(self.path,i,self.job_string))

                #read the excitonic wavefunction
                if Format == 'x':
                    ewf = YamboExcitonWaveFunctionXSF()
                else:
                    ewf = YamboExcitonWaveFunctionGnuplot()
                ewf.read_file("%s/%s"%(self.path,filename))
                data = ewf.get_data()
                for word in keywords:
                    if word in data:
                        self.data[word] = data[word]

                #calculate center of mass of atoms
                lat = np.array(data["lattice"])
                center_atom = np.zeros([3])
                for atype,x,y,z in data["atoms"]:
                    center_atom += np.array([x,y,z])
                center_atom /= len(data["atoms"])
                center_atom_red = car_red([center_atom],lat)[0]

                #shift wavefunctions grid to center of mass
                nx = data['nx'] 
                ny = data['ny'] 
                nz = data['nz']

                #make center_atom_red commensurate with fft
                center_atom_red = center_atom_red * np.array([nx,ny,nz])
                center_atom_red_int = [int(x) for x in center_atom_red]
                displacement = np.array([nx,ny,nz])/2-center_atom_red_int
                dx,dy,dz = displacement

                # shift grid
                # http://www.xcrysden.org/doc/XSF.html 
                dg  = np.array(data["datagrid"]).reshape([nz,ny,nx])
                dg  = np.roll(dg,dx,axis=2)
                dg  = np.roll(dg,dy,axis=1)
                dg  = np.roll(dg,dz,axis=0)
                data["datagrid"] = dg.flatten()

                #shift atoms
                atoms = []
                dx,dy,dz = red_car([displacement/np.array([nx,ny,nz],dtype=float)],lat)[0]
                for atype,x,y,z in data["atoms"]:
                    atoms.append([atype,x+dx,y+dy,z+dz])
                self.data["atoms"] = atoms

            ##############################################################
            # Excitonic Amplitudes
            ##############################################################
            #create ypp input for the amplitudes file and run
            yppew["States"] = "%d - %d"%(i,i)
            yppew.write("%s/yppew_%d.in"%(self.path,i))

            filename = "%s/o-%s.exc_weights_at_%d"%(self.path,self.job_string,i)
            if not os.path.isfile(filename):
                os.system("cd %s; ypp -F yppew_%d.in -J %s"%(self.path,i,self.job_string))

            #read the excitonic weigths
            ew = YamboExcitonWeight(filename,save=self.save,path=self.path)
            qpts, weights = ew.calc_kpts_weights(repx=repx,repy=repy,repz=repz)

            ############
            # Save data
            ############
            exciton = {"energy": e,
                       "intensity": intensity,
                       "weights": weights,
                       "qpts": qpts,
                       "index": i}
            if wf:
                exciton["hole"] = Hole
                exciton["datagrid"] = np.array(data["datagrid"])

            self.data["excitons"].append(exciton)

    def plot(self):
        pass

    def write_json(self,filename="absorptionspectra",verbose=0):
        """ 
        Write a jsonfile with the absorption spectra 
        and the wavefunctions of certain excitons
        """
        JsonDumper(self.data,"%s.json"%filename)

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app(str(self.excitondb.get_string(mark='-')))
        return "\n".join(lines)



# Copyright (c) 2018, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
from yambopy import *
from yambopy.plot import *
import os

ha2ev = 27.211396132

class YamboRTDB(object):
    """
    Open the RT databases and store it in a RTDB class.
    So far treating COHERENT runs: reading polarization and carriers.
    """

    def __init__(self,folder='.',calc='.',observablesdb='ndb.RT_OBSERVABLES',carriersdb='ndb.output_carriers'):
        # Find path with RT data
        obspath = '%s/%s/%s'%(folder,calc,observablesdb)
        carrpath = '%s/%s/%s'%(folder,calc,carriersdb)

        try:
            data_obs= Dataset(obspath)
        except:
            raise ValueError("Error reading OBSERVABLES database at %s"%obspath)

        try:
            data_carr= Dataset(carrpath)
        except:
            raise ValueError("Error reading CARRIERS database at %s"%carrpath)


        self.read_observables(data_obs)
        self.read_carriers(data_carr)

        data_obs.close()
        data_carr.close()

    def read_observables(self,database):
        """
        Read polarization
        """
        pol = database.variables['Polarization']
        pol = np.array(pol).T
        pol = pol[0] + 1j*pol[1]
        self.polarization = pol.real #Real part only

    def read_carriers(self,database):
        """
        Read output carriers
        """
        self.diff_carriers = database.variables['carriers_values'][:,1]    
        self.hole_carriers = database.variables['carriers_values'][:,2]
        self.elec_carriers = database.variables['carriers_values'][:,3]
        #Ratio between carrier difference and average
        self.ratio_carriers = self.diff_carriers/(0.5*(self.hole_carriers+self.elec_carriers))
    
    def __str__(self):
        s = ""
        s += "times: %d\n"%self.polarization.shape[1]
        return s




#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: RR
#
# This file is part of the yambopy project
#
from yambopy.io.xsffile import YamboXsf

def interlayer_power(xsf_hole_bottom_file, xsf_hole_top_file, zthreshold, c , fractional = False):
    xsf_hole_bottom = YamboXsf.read_xsf(xsf_hole_bottom_file)
    xsf_hole_top = YamboXsf.read_xsf(xsf_hole_top_file)        
    b_hole_b_contr, t_hole_b_contr = xsf_hole_bottom.contribution_twolayers(zthreshold,c, fractional )
    b_hole_t_contr, t_hole_t_contr = xsf_hole_top.contribution_twolayers(zthreshold,c, fractional )
    print('ciao',t_hole_b_contr+b_hole_b_contr+b_hole_t_contr+t_hole_t_contr)
    return ((t_hole_b_contr+b_hole_t_contr)/(b_hole_b_contr+t_hole_b_contr+t_hole_t_contr+b_hole_t_contr))



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
#
import os

def get_pseudo_path(pseudo_name,absolute_paths):
    """Check if a pseudo with pseudo_name exists in any of the paths"""
    for path in absolute_paths:
        for filename in os.listdir(path):
            if pseudo_name == filename:
                return os.path.join(path,filename)
    return None



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: RR
#
# This file is part of the yambopy project
#
import re
import numpy as np
from yambopy.dbs.latticedb import *

Bohr2Ang = 0.529177249

class YamboXsf():
    def __init__(self):
        self.cell_parameters = None
        self.cell_parameters_conv= None
        self.l_conventional = False
        self.dim = None
        self.natoms = None
        self.atom_positions = []
        self.grid_data = []
        self.data_array = np.array([])
        self.lattice = None
        self.grid_dim = np.array([])

    def set_cell_parameters(cls, cell_parameters):
        cls.cell_parameters = cell_parameters
    
    def set_cell_parameters_conv(cls, cell_parameters_conv):
        cls.cell_parameters_conv = cell_parameters_conv     

    def add_atom(cls, symbol, position):
        cls.atom_positions.append((symbol, position))

    def set_dim(cls,dim='3D'):
        cls.dim = dim
    
    def set_natoms(cls,natoms):
        cls.natoms = natoms        
        
    def set_lconventional(cls, l_conv):
        cls.l_conventional = l_conv
    
    def add_grid_data(cls, grid_name, block_dim, sub_grid_name, sub_grid_size, sub_grid_dim, sub_grid_origin,\
        sub_grid_vectors, data_array):
        cls.grid_data.append((grid_name, block_dim, sub_grid_name, sub_grid_size, sub_grid_dim, sub_grid_origin, \
            sub_grid_vectors, data_array))     

    def get_data_array(cls, data_array):
        cls.data_array = np.transpose(data_array,axes=(2,1,0)) # I want it in python order, to be checked if it's ok  

    def get_grid_dim(cls, grid_dim):
        cls.grid_dim = grid_dim

    def set_lattice(cls,lattice):
        cls.lattice = lattice

    def write_xsf(self, filename):
        with open(filename, 'w') as f:
            if (self.dim==3) : f.write("CRYSTAL\n")
            if (self.dim==2) : f.write("SLAB\n")
            if (self.dim==1) : f.write("POLYMER\n")
            if (self.dim==0) : f.write("MOLECULE\n")
            #PRIMVEC
            if np.any(self.cell_parameters):
                f.write("PRIMVEC\n")
                for vector in self.cell_parameters:
                    f.write("  ".join(str(x) for x in vector) + "\n")
            #CONVVEC
            if np.any(self.cell_parameters_conv):
                f.write("CONVVEC\n")
                for vector in self.cell_parameters_conv:
                    f.write("  ".join(str(x) for x in vector) + "\n")
            #ATOM_POSITIONS
            if (len(self.atom_positions) > 0):
                f.write("PRIMCOORD\n")
                f.write(f"{len(self.atom_positions)} 1\n")
                for symbol, position in self.atom_positions:
                    f.write(f"{symbol} {position[0]} {position[1]} {position[2]}\n")

        # wride grid_data for each sub_block
            if (len(self.grid_data)!=0):
                new_block_counter = 0 
                for grid_name, block_dim, sub_grid_name, sub_grid_size, sub_grid_dim, sub_grid_origin, sub_grid_vectors, sub_data_array in self.grid_data:
                    if(new_block_counter == len(sub_grid_name) or new_block_counter==0): f.write(f"BEGIN_BLOCK_DATAGRID_{block_dim}D\n")
                    f.write(f"\t{grid_name}\n")
                    f.write(f"\tBEGIN_DATAGRID_{sub_grid_dim}D\n")
                    f.write(f"".join("\t"+str(sub_grid_size[i]) for i in range(len(sub_grid_size)))+"\n")
                    f.write(f"".join("\t"+str(sub_grid_origin[i]) for i in range(len(sub_grid_origin)))+"\n")
                    for i in range(0,sub_grid_vectors.shape[0]):
                        f.write(f"".join("\t"+str(sub_grid_vectors[i][j]) for j in range(sub_grid_vectors.shape[1]))+"\n")
                    for i in range(sub_data_array.shape[0]):
                        for j in range(sub_data_array.shape[1]):
                            for k in range(sub_data_array.shape[2]):
                                #print(sub_data_array[i,j,k]) 
                                f.write(f"\t{sub_data_array[i,j,k]}\n")
                    f.write(f"\tEND_DATAGRID_{sub_grid_dim}D\n")
                    new_block_counter+=1
                    if(new_block_counter == len(sub_grid_name)or new_block_counter-1==0): f.write(f"END_BLOCK_DATAGRID_{block_dim}D")

    @staticmethod
    def read_xsf(filename):
        xsf_file = YamboXsf()

        with open(filename, 'r') as f:
            lines = [line.lstrip() for line in f.readlines()]

        #extract the dimension
        try: 
            if(lines.index("CRYSTAL\n")) : xsf_file.set_dim(3) 
            elif(lines.index("SLAB\n")) : xsf_file.set_dim(2)
            elif(lines.index("POLYMER\n")) : xsf_file.set_dim(1)
            elif(lines.index("MOLECULE\n")) : xsf_file.set_dim(0)
            else:
                ValueError("xsf file does not contain either CRYSTAL, SLAB, POLYMER, or MOLECULE")
        except ValueError as e:
            print(e)
        # Extract cell parameters
        cell_parameters = []
        cell_parameters_start = lines.index("PRIMVEC\n") + 1
        cell_parameters_end = cell_parameters_start + 3
        for line in lines[cell_parameters_start:cell_parameters_end]:
            cell_parameters.append([float(x) for x in line.split()])
        #Needed only if the convention cell is specified
        try:
            if(lines.index("CONVVEC\n")):
                xsf_file.set_lconventional(True)
                cell_parameters_conv=[]
                cell_parameters_conv_start = lines.index("CONVVEC\n")+1
                cell_parameters_conv_end = cell_parameters_conv_start+3
                for line in lines[cell_parameters_conv_start:cell_parameters_conv_end]:
                    cell_parameters_conv.append([float(x) for x in line.split()])    
                xsf_file.set_cell_parameters_conv(cell_parameters_conv)
            else:
                ValueError("conventional cell not present or not found.")
        except ValueError as e:       
            print(e) 
        # primitive cell parameters 
        xsf_file.set_cell_parameters(cell_parameters)
        # Extract atom positions
        atom_positions_start = lines.index("PRIMCOORD\n") + 1
        natoms = int(lines[atom_positions_start].split()[0]) # read the number of atoms
        atoms_positions_end = atom_positions_start +natoms +1
        for line in lines[atom_positions_start+1:atoms_positions_end]:
            symbol, x, y, z = line.split()
            position = [float(x), float(y), float(z)]
            xsf_file.add_atom(symbol, position)

        # Extract grid data
        block_start_idx = []
        block_dim = []
        
        for i,line in enumerate(lines): 
            if line.startswith(f"BEGIN_BLOCK_DATAGRID_"):
                block_start_idx.append(i)  
                block_dim.append(int(re.findall('[0-9]',line)[0])) 
                print(block_start_idx, block_dim)

        for istart_idx, start_idx in enumerate(block_start_idx):
            print(istart_idx,start_idx)
            block_end_idx = lines.index(f"END_BLOCK_DATAGRID_{block_dim[istart_idx]}D\n", start_idx) + 1
            block_lines = lines[start_idx:block_end_idx]
            #print(block_lines)

            subblock_start_idx = [i for i,line in enumerate(lines) if line.startswith(f"BEGIN_DATAGRID_{block_dim[istart_idx]}D\n")]
            print(subblock_start_idx)
            grid_name = block_lines[1].strip()
            for sub_start_idx in subblock_start_idx:
                sub_data_vectors=[]
                sub_block_end_idx = lines.index(f"END_DATAGRID_{block_dim[istart_idx]}D\n", sub_start_idx)+1
                sub_block_lines = lines[sub_start_idx:sub_block_end_idx]
                sub_block_name = re.split('_',sub_block_lines[0]) # extract name of the subblock
                sub_data_size = [int(x) for x in sub_block_lines[1].split()]
                sub_data_num = len(sub_data_size) # number of subdata for each block
                sub_data_origin = sub_block_lines[2].split()
                for j in range(3, 3+ sub_data_num):
                    sub_data_vectors.append([float(x) for x in sub_block_lines[j].split()])
                sub_data_values = []
                for j in range(3+sub_data_num,len(sub_block_lines)-1): # -1 becuase I do not want to include the last line
                    sub_data_values.append([float(x) for x in sub_block_lines[j].split()])
                sub_data_array = np.array(sub_data_values).reshape(np.flip(sub_data_size)) #reverse ordering cause fortran
                if(block_dim[istart_idx]==2): 
                    na = np.newaxis
                    sub_data_array = sub_data_array[:,:,na]
                xsf_file.get_data_array(data_array=sub_data_array)
                xsf_file.get_grid_dim(sub_data_size)
                xsf_file.add_grid_data(grid_name,block_dim=block_dim[istart_idx],sub_grid_name=sub_block_name, sub_grid_size=sub_data_size, \
                    sub_grid_dim=block_dim[istart_idx], sub_grid_origin=sub_data_origin, \
                    sub_grid_vectors=np.array(sub_data_vectors), data_array=sub_data_array)

        return xsf_file
    #instance of xsf file from lattice db class
    @staticmethod
    def from_latticedb(yambolattice):
        xsf_file = YamboXsf()
        xsf_file.set_dim(3) # I am not aware of Yambopy working with 2D 1D or 0D structures, for now
        xsf_file.set_lattice(yambolattice)
        xsf_file.set_cell_parameters(xsf_file.lattice.lat*Bohr2Ang)
        for iatom in range(0,len(xsf_file.lattice.car_atomic_positions)):
            xsf_file.add_atom(xsf_file.lattice.atomic_numbers[iatom],xsf_file.lattice.car_atomic_positions[iatom]*Bohr2Ang)
        return xsf_file
    
    def contribution_twolayers(self, zthreshold, c , fractional = True):
        data = self.data_array
        nGx, nGy, nGz = data.shape # G-vectors in real space
        # Now I need to know the range of G-vector in z that are < zthreshold or above
        # assume zthreshold is in fractionalcoordinates
        if (fractional):
            nGz_bottom = int(np.floor(nGz*(zthreshold)))
            bottom_layer = np.sum(data[:,:,:nGz_bottom])
            top_layer = np.sum(data[:,:,nGz_bottom:-1])
            return bottom_layer, top_layer
        else:
            zthreshold = zthreshold/c
            nGz_bottom = int(np.floor(nGz*(zthreshold)))
            print('ngz', nGz_bottom)
            bottom_layer = np.sum(data[:,:,:nGz_bottom])
            top_layer = np.sum(data[:,:,nGz_bottom:-1])
            return bottom_layer, top_layer



# Author = Riccardo Reho r.reho@uu.nl
#
# This file is part of yambopy
#
# Interface with pp.x, post-processing tool of QE

import os
import re
import shutil
import numpy as np
from qepy import qepyenv

class PPIn():
    """
    Class to generate an manipulate Quantum Espresso post-processing pp.x input files
    This class is not meant to be comprehensive but a lightweight version 
    capable of basic input/ouput of PW files.

    Examples of use:

    To write a local file from scratch with name "pp.in"

        .. code-block :: python

            pp = PwIn.from_file('prefix.pp')
            print pp

    For now, this has been tested only for the plotting of wavefunctions |psi**2| from QE, as
    in the followin example.

        .. code-block :: python

            from yambopy import *
            prefix = 'prefix'
            kpoints =np.arange(1,2,1)
            kbands = np.arange(456,457,1)
            nf = (kbands[-1]-kbands[0]+1)*len(kpoints) #+1 due to the usual python counting
            print ("Post-processing pp.x\n nfile =", nf,"\n")

            files = []
            weights=np.ones(nf)
            for i_k, k in enumerate(kpoints):
                for i_b,b in enumerate(kbands):
                    files.append(f"'./wfc-datfiles/{prefix}_K{str(k).zfill(3)}_B{str(b).zfill(3)}'")
            # Initialize PPIn class
            pp = PPIn()
            # In this specific case, the .dat files are already available from previous run, therefore I only need to plot
            pp.set_nfile(nf)
            pp.set_files(files)
            pp.set_weigths(weights)
            pp.set_fileout(f"'plot-B{kbands[-1]}-B{kbands[0]}-K{kpoints[-1]}-K{kpoints[0]}.xsf'")
            pp.write(f'plot-B{kbands[0]}-B{kbands[-1]}-K{kpoints[0]}-K{kpoints[-1]}.in',plotonly=True)

    """

    _pp = "pp.x"

    def __init__(self):
        """ TODO: specify the required parameters """
        #dictionaries
        self.inputpp = dict()
        self.plot = dict(nfile = 1, output_format=5, fileout = "'plot.xsf'", iflag=3)

    def set_prefix(self,prefix):
        self.inputpp['prefix'] = prefix
    
    def set_outdir(self,outdir):
        self.inputpp['outdir'] = outdir

    def set_filplot(self,filplot):
        self.inputpp['filplot'] = filplot

    def set_plotnum(self,plotnum):
        self.inputpp['plotnum'] = plotnum

    def set_kpoints(self,kpoints):
        "k-point should be an array with a range of values you want to compute the quantities or a single k-point"
        if (len(kpoints)==1):
            if (isinstance(kpoints,int)):
                self.inputpp['kpoint(1)'] = kpoints
            if (isinstance(kpoints,list)):
                self.inputpp['kpoints(1)'] = kpoints[0]
        else: 
            if(len(kpoints) >1):
                self.inputpp['kpoints(1)'] = kpoints[0]
                self.inputpp['kpoints(2)'] = kpoints[1]

    def set_kbands(self,kbands):
        "k-point should be an array with a range of values you want to compute the quantities or a single k-point"
        if (len(kbands)==1):
            if (isinstance(kbands,int)):
                self.inputpp['kbands(1)'] = kbands
            if (isinstance(kbands,list)):
                self.inputpp['kbands(1)'] = kbands[0]
        else:
            if(len(kbands) >1):
                self.inputpp['kbands(1)'] = kbands[0]
                self.inputpp['kbands(2)'] = kbands[1]

    def set_nfile(self,nfile):
        self.plot['nfile'] = nfile

    def set_files(self,files):
        for i_f,f in enumerate(files):
            self.plot[f'filepp({i_f+1})'] = f #+1 for python-Fortran counting
    
    def set_weigths(self,weights):
        for i_w,w in enumerate(weights):
            self.plot[f'weight({i_w+1})'] = w #+1 for python-Fortran counting

    def set_outputformat(self,outputformat):
        self.plot['output_format'] = outputformat
    
    def set_fileout(self,fileout):
        self.plot['fileout'] = fileout

    def set_iflag(self,iflag):
       self.plot['iflag'] = iflag
       
    def stringify_group(self, keyword, group):
        if group != {}:
            string='&%s\n' % keyword
            for keyword in sorted(group): # Py2/3 discrepancy in keyword order
                string += "    %s = %s\n" % (keyword, group[keyword]) #4 space are mandatory at the beg of the line for pp.x syntax
            string += "/"
            return string
        else:
            return ''
    
    def write(self,filename,plotonly=False):
        """write the file to disk """
        with open(filename,'w') as f:
           if (plotonly):
              """an empyt block for inputpp is mandatory in QE if you only want to plot results from data"""
              f.write('&inputpp \n/\n')
           f.write(str(self))               
    def get_string(self):
        """
        Output the file in the form of a string
        """
        lines = []; app = lines.append
        app( self.stringify_group("inputpp",self.inputpp) ) #print inputpp
        app( self.stringify_group("plot",self.plot) ) #print plot
    
        return "\n".join(lines)

    def __str__(self):
        return self.get_string()    



from __future__ import print_function
# Version : dec 2nd
from builtins import str
from builtins import range
import os
import re
import numpy as np

# Get a list of all exciton files, make a new file with x=time, y1 = exc A1, y2 = exc A2

# List of exciton files
dirfiles = os.listdir('.')
files=[]
for f in dirfiles:
    if '_E_' in f:
        files.append(f)

files.sort()
print(files)

# Get times
times=[]
for f in files:
    time=re.search('[0-9]{3,}',f).group()
    times.append(time)

# Get excitons
excitons=[]
i_thr=0.005 # intensity threshold
exc_list = (1,2) # excitons wanted (here 1st and 2nd one that respect the threshold)

for p,f in enumerate(files):
    n=1
    print(files[p],times[p])
    content = np.loadtxt(f)
    for e,i,index in content:
        if i>i_thr:
            if n in exc_list:
                n=n+1
                excitons.append((times[p],e))
# excitons is an array of the form
# excitons[(time0,energy_n),(time0,energy_n+1),...,(timep,energy_n+m)]


f = open('output.dat','w')
for i in range(0,len(excitons)):
    t = excitons[i][0]
    e = excitons[i][1]
    line = str(t)+'\t'+str(e)+'\n'
    f.write(line)

f.close()



# Copyright (c) 2018, Fulvio Paleari, Alejandro Molina-Sánchez, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
from yambopy import *
from netCDF4 import Dataset
from math import sqrt
import numpy as np
from yambopy.tools.string import marquee
import os
import matplotlib.pyplot as plt
from yambopy.units import ha2ev, I
from yambopy.plot.plotting import add_fig_kwargs,BZ_hexagon,shifted_grids_2D

class YamboExcitonPhononDB():
    """
    Python class to read the exciton-phonon matrix elements from yambo.

    By default it reads the full databases including fragments.

    - Input: YamboLatticeDB object
    - Input: paths of ndb.excph*
    - Input: if not read_all, read only header

    NB: So far we read G(exc_in_Q=0, exc_out_q, ph_q) 

    - Usage and main variables:

      :: yexcph = YamboExcitonPhononDB(ylat,save_excph=path1)

      :: yexcph.car_qpoints  #Momenta cc, iq
      :: yexcph.red_qpoints  #Momenta rlu, iq
      :: yexcph.exciton_in   #Exciton states in, iexc1
      :: yexcph.exciton_out  #Exciton states out (summed in self-energy), iexc2
      :: yexcph.phonon_modes #Phonon modes, il
      :: yexcph.excph        #Exc-ph matrix elements
      :: yexcph.excph_sq     #Exc-ph matrix elements squared

    Formats:
    - excph[iq][il][iexc1][iexc2]

    Plots provided:
    - Call function plot for scatterplot in the q-BZ of any quantity A(q)_{inu,ib1,ib2}
        -- if plt_show, show plot at runtime
        -- if plt_cbar, add colorbar

      Example, plot of |G(q)_{3,4,4}|:
           :: yexcph.plot_excph( np.abs(yexcph.excph[:,3,4,4]) )
    """
    def __init__(self,lattice,save_excph="./",read_all=True):
        
        # Find databases
        if os.path.isfile("%s/ndb.excph_gkkp"%save_excph): filename='%s/ndb.excph_gkkp'%save_excph
        else: raise FileNotFoundError("Databases not found.")
        self.frag_filename = filename + "_fragment_"
                    
        # Check if databases exist. Exit only if header is missing.
        try: database = Dataset(filename)
        except: raise FileNotFoundError("error opening %s in YamboElectronPhononDB"%filename)
        
        try: database_frag = Dataset("%s1"%self.frag_filename)
        except FileNotFoundError: print("[WARNING] Database fragment at q=0 not detected")
        else: database_frag.close()
                 
        #read dimensions of exciton phonon parameters
        self.nexc_i = database.variables['EXCITON_STATES'][1].astype(int)
        self.nexc_o = database.variables['EXCITON_SUM'][1].astype(int)
        self.nmodes = database.variables['PHONON_MODES'][0].astype(int)
        self.nqpoints = database.variables['HEAD_R_LATT'][3].astype(int)
        self.type_exc_i = database.variables['L_kind_in'][...].tostring().decode().strip()
        self.type_exc_o = database.variables['L_kind_out'][...].tostring().decode().strip()
        database.close()

        #Check how many databases are present
        self.nfrags = self.nqpoints
        for iq in range(self.nqpoints):
            if not os.path.isfile("%s%d"%(self.frag_filename,iq+1)): 
                self.nfrags = iq
                break
        
        # Necessary lattice information
        self.lattice = lattice
        self.alat    = lattice.alat
        self.rlat    = lattice.rlat
        
        # Keep reading
        if read_all: self.read_full_DB()

    def read_full_DB(self):
        """
        Read more variables in the ndb.excph_gkkp_fragments* dbs as attributes of this class
        """
        
        # qpoints
        self.read_qpoints()
        
        # Matrix elements
        self.read_excph()
        
    def read_qpoints(self):
        """
        Read q points and return cartesian and reduced coordinates
        """
        var_nm = "EXCPH_Q"
        
        self.qpoints = np.zeros([self.nfrags,3])

        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil)
            self.qpoints[iq] = database.variables['%s%d'%(var_nm,iq+1)][:].T        
    
        self.car_qpoints = np.array([ q/self.alat for q in self.qpoints ])

    def read_excph(self):
        """
        Read exciton-phonon matrix elements and their modulus squared
        
        NB: EXCPH_GKKP_Q is saved by yambo as (2,mode,exc_out,exc_in), but netCDF stores
            the *transpose* (exc_in,exc_out,mode,2).
            We want to change it to complex (iq,mode,exc_in,exc_out)
        """    
        var_nm    = "EXCITON_PH_GKKP_Q"
        var_sq_nm = "EXCITON_PH_GKKP_SQUARED_Q"
            
        # excph[q][mode][iexc1][iexc2]
        excph_full    = np.zeros([self.nfrags,self.nmodes,self.nexc_i,self.nexc_o],dtype=np.complex64)  
        excph_sq_full = np.zeros([self.nfrags,self.nmodes,self.nexc_i,self.nexc_o])  
        
        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil)
            excph = database.variables['%s%d'%(var_nm,iq+1)][:]
            excph_full[iq] = np.moveaxis( excph[:,:,:,0]+I*excph[:,:,:,1], -1,0 )
            #excph_full[iq] = np.swapaxes( np.swapaxes(excph[:,:,:,0] + I*excph[:,:,:,1],-1,0), -1,-2)
            
            excph_sq = database.variables['%s%d'%(var_sq_nm,iq+1)][:]
            excph_sq_full[iq] = np.moveaxis( excph_sq, -1,0)
            #excph_sq_full[iq] = np.swapaxes( np.swapaxes(excph_sq[:,:,:],-1,0), -1,-2)
            database.close()
        
        # Check integrity of elph values
        if np.isnan(excph_full).any(): print('[WARNING] NaN values detected in elph database.')

        self.excph    = excph_full
        self.excph_sq = excph_sq_full

    @add_fig_kwargs
    def plot_excph(self,data,plt_show=False,plt_cbar=False,**kwargs):
        """
        2D scatterplot in the q-BZ of the quantity A_{iq}(ib2,ib1,inu).
        
        Any real quantity which is a function of only the q-grid may be supplied.
        The indices iq,inu,ib1,ib2 are user-specified.
        
        - if plt_show plot is shown
        - if plt_cbar colorbar is shown
        - kwargs example: marker='H', s=300, cmap='viridis', etc.
        
        NB: So far requires a 2D system. 
            Can be improved to plot BZ planes at constant k_z for 3D systems.
        """
              
        qpts = self.car_qpoints
        
        # Input check
        if len(data)!=len(qpts): 
            raise ValueError('Something wrong in data dimensions (%d data vs %d kpts)'%(len(data),len(kpts)))
        
        # Global plot stuff
        self.fig, self.ax = plt.subplots(1, 1)
        self.ax.add_patch(BZ_Wigner_Seitz(self.lattice))
        
        if plt_cbar:
            if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
            else:                       color_map = plt.get_cmap('viridis') 
        lim = 1.05*np.linalg.norm(self.rlat[0])
        self.ax.set_xlim(-lim,lim)
        self.ax.set_ylim(-lim,lim)

        # Reproduce plot also in adjacent BZs
        BZs = shifted_grids_2D(qpts,self.rlat)
        for qpts_s in BZs: plot=self.ax.scatter(qpts_s[:,0],qpts_s[:,1],c=data,**kwargs)
        
        if plt_cbar: self.cbar = self.fig.colorbar(plot)
        
        plt.gca().set_aspect('equal')

        if plt_show: plt.show()
        else: print("Plot ready.\nYou can customise adding savefig, title, labels, text, show, etc...")

    def __str__(self,verbose=False):

        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
            
        app('nqpoints: %d'%self.nqpoints)
        app('nmodes: %d'%self.nmodes)
        app('nexcitons in (L_%s): %d'%(self.type_exc_i,self.nexc_i))
        app('nexcitons out (L_%s): %d'%(self.type_exc_o,self.nexc_o))
        if self.nfrags == self.nqpoints: app('fragments: %d'%self.nfrags)
        else: app('fragments: %d [WARNING] nfrags < nqpoints'%self.nfrags)
        
        return "\n".join(lines)



import os
from netCDF4 import Dataset
from yambopy import *
from yambocommandline.commands import generate_save
from schedulerpy import *
import argparse

"""
Script to produce a yambo SAVE folder along with gkkp matrix elements.
Usable to quickly initialize a yambo_ph (equilibrium) calculation.

Inputs:
 1. --nscf_dir='path/to/nscf/save/folder' [OPTIONAL]
 2. --elph_dir='path/to/dfpt/elph_dir/folder'
 3. --yambo='path/to/yambo/executables' [OPTIONAL]
 4. --expand                            [OPTIONAL]

NB:
 The SAVE folder is created in the directory where the script is launched!

"""
def get_q_k_size(database,elph_save):
    """
    Get number of k and q points in the (IBZ) grids
    """
    # kpoints
    db = Dataset(database+"/SAVE/ns.db1")
    Nk = len(db.variables['K-POINTS'][:].T)
    db.close()
    # qpoints
    Nq = len(glob('./elph_dir/s.dbph_0*'))
    return Nq,Nk

def generate_gkkp(database,qe_save,elph_save,y_dir,expand,scheduler):
    """
    Read gkkp from dfpt calculation
    """

    def run_ypp_ph(UseQindxB=False):
        """
        Run ypp_ph and do checks
        """
        yph = YamboIn.from_runlevel('-gkkp',executable=ypp_ph,filename=filnm2,folder=database)
        # Apparently, now gkkp_db must be always specified
        yph.arguments.append('gkkp_db')
        if expand:
            #if Nq!=Nk: yph.arguments.append('gkkp_db')
            yph.arguments.append('GkkpExpand')
            if UseQindxB: yph.arguments.append('UseQindxB')
            print('    expanding gkkp in the full BZ')
        yph['DBsPATH'] = "./elph_dir"
        if os.path.isfile('%s/s.dbph_bare_000001'%elph_save):
            print('    reading also bare gkkp')
            yph.arguments.append('GkkpReadBare')
        yph.write('%s/%s'%(database,filnm2)) 

        yppph_run = scheduler()
        yppph_run.add_command('cd %s ; %s -F %s; cd -'%(database,ypp_ph,filnm2))
        yppph_run.run()

    def dbs_are_not_there():
        """
        Check if elph databases were produced succesfully
        """
        return ( not os.path.isfile('%s/SAVE/ndb.elph_gkkp'%database) ) and \
               ( not os.path.isfile('%s/SAVE/ndb.elph_gkkp_expanded'%database) )

    # Generate SAVE folder if nscf_path is given
    if qe_save != "": generate_save.generate_save(database,qe_save,y_dir,scheduler,noinit=True)
    
    #check if gkkp databases are already present
    if os.path.isfile('%s/SAVE/ndb.elph_gkkp'%database) or os.path.isfile('%s/SAVE/ndb.elph_gkkp_expanded'%database):
        print('gkkp found!')
    else:
        #check if the elph_dir folder is present
        if not os.path.isfile('%s/s.dbph_000001'%elph_save):
            print('problem with dbph databases at %s'%elph_save)
            exit()
        else:
            print('reading gkkp')

            if y_dir!="": y_dir=y_dir+"/"
            yambo_ph = y_dir+"yambo_ph"
            ypp_ph = y_dir+"ypp_ph"
            filnm1 = 'setup.in'
            filnm2 = 'gkkp.in'

            y1 = YamboIn.from_runlevel('-i -V RL',executable=yambo_ph,filename=filnm1,folder=database)
            y1.arguments.append('BSEscatt')
            y1['K_grids'] = "BSC"
            y1.write('%s/%s'%(database,filnm1))
            yamboph_run = scheduler()
            if not os.path.islink('%s/elph_dir'%database): yamboph_run.add_command('cd %s ; ln -s %s . ; cd -'%(database,elph_save))
            # Check if a custom grid is used
            Nk,Nq = get_q_k_size(database,elph_save)
            # Regular grid
            if Nk==Nq: yamboph_run.add_command('cd %s ; %s -F %s -J ./elph_dir ; cd -'%(database,yambo_ph,filnm1))
            # Custom grid
            else: yamboph_run.add_command('cd %s ; %s -F %s ; cd -'%(database,yambo_ph,filnm1))
            yamboph_run.run()

            # Run ypp_ph
            run_ypp_ph()
            
            dbs_are_not_there = ( not os.path.isfile('%s/SAVE/ndb.elph_gkkp'%database) ) and \
                                ( not os.path.isfile('%s/SAVE/ndb.elph_gkkp_expanded'%database) )
            if dbs_are_not_there:
                print("[WARNING] First attempt didn't work. Retrying with UseQindxB")
                run_ypp_ph(UseQindxB=True)
                #if dbs_are_not_there(): print('[ERROR] ndb.elph_gkkp databases not created. Check the logs.')
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Generate SAVE folder including gkkp databases')
    parser.add_argument('-nscf','--nscf_dir', type=str, default="", help='<Optional> Path to nscf save folder')
    parser.add_argument('-elph','--elph_dir', type=str,help='<Required> Path to elph_dir folder',required=True)
    parser.add_argument('-y','--yambo_dir', type=str, default="", help='<Optional> Path to yambo executables')
    parser.add_argument('-e','--expand', action="store_true", help="Expand GKKP")
    args = parser.parse_args()

    nscf_dir = args.nscf_dir
    elph_dir = args.elph_dir
    yambo_dir = args.yambo_dir
    expand = args.expand

    database = './'
    scheduler = Scheduler.factory

    generate_gkkp(database,nscf_dir,elph_dir,yambo_dir,expand,scheduler)



from __future__ import print_function
from builtins import str
from builtins import range
import matplotlib
#matplotlib.use('Agg') # prevents crashes if no X server present (clusters)
from yambopy import *
import matplotlib.pyplot as plt
import sys
import argparse
import numpy as np
import operator

"""
Study the convergence of GW calculations by looking at the change in band-gap value.

The script reads from <folder> all results from <variable> calculations and display them.

Use the band and k-point options (or change default values) according to the size of your k-grid and
the location of the band extrema.
"""

parser = argparse.ArgumentParser(description='Study GW convergence with regards to the band-gap value.')
parser.add_argument('folder'    , help='Folder containing SAVE and convergence runs.')
parser.add_argument('variable'  , help='Variable tested (e.g. FFTGvecs)'             )
parser.add_argument('-bc','--bandc'     , help='Lowest conduction band number'    , default=53, type=int)
parser.add_argument('-kc','--kpointc'   , help='K-point index for conduction band', default=19, type=int)
parser.add_argument('-bv','--bandv'     , help='Highest valence band number'      , default=52, type=int)
parser.add_argument('-kv','--kpointv'   , help='K-point index for valence band'   , default=1, type=int)
parser.add_argument('-np','--nopack'    , help='Skips packing o- files into .json files', action='store_false')
parser.add_argument('-t'  ,'--text'      , help='Also print a text file for reference'   , action='store_true')
args = parser.parse_args()

folder = args.folder
var    = args.variable
bandc  = args.bandc
kpointc= args.kpointc
bandv  = args.bandv
kpointv= args.kpointv
nopack = args.nopack
text   = args.text

print('Valence band: ',bandv,'conduction band: ',bandc)
print('K-point VB: ',kpointv, ' k-point CB: ',kpointc)


# Packing results (o-* files) from the calculations into yambopy-friendly .json files
if nopack: # True by default, False if -np used
    print('Packing ...')
    pack_files_in_folder(folder,mask=var)
    pack_files_in_folder(folder,mask='reference')
    print('Packing done.')
else:
    print('Packing skipped.')

# importing data from .json files in <folder>
print('Importing...')
data = YamboAnalyser(folder)

# extract data according to relevant variable
outvars = data.get_data(var)
invars = data.get_inputfiles_tag(var)
tags = data.get_tags(var)

# Get only files related to the convergence study of the variable,
# ordered to have a smooth plot
keys=[]
sorted_invars = sorted(list(invars.items()), key=operator.itemgetter(1))

for i in range(0,len(sorted_invars)):
    key=sorted_invars[i][0]
    if key.startswith(var) or key=='reference.json':
        keys.append(key)
print('Files detected: ',keys)

print('Preparing output...')
### Output

# Unit of the variable :
unit = invars[keys[0]]['variables'][var][1]

# The following variables are used to make the script compatible with both short and extended output
#kpindex = tags[keys[0]].tolist().index('K-point')
kpindex = tags[keys[0]].tolist().index('Kpoint_index')  # Alejandro
bdindex = tags[keys[0]].tolist().index('Band')
e0index = tags[keys[0]].tolist().index('Eo')
gwindex = tags[keys[0]].tolist().index('E-Eo')


array = np.zeros((len(keys),2))

for i,key in enumerate(keys):
    # input value
    # GbndRnge and BndsRnX_ are special cases
    if var.startswith('GbndRng') or var.startswith('BndsRnX'):
        # format : [1, nband, ...]
        array[i][0] = invars[key]['variables'][var][0][1]
    else:
        array[i][0] = invars[key]['variables'][var][0]

    # Output value (gap energy)
    # First the relevant lines are identified
    valence=[]
    conduction=[]
    for j in range(len(outvars[key]+1)):
        if outvars[key][j][kpindex]==kpointc and outvars[key][j][bdindex]==bandc:
                conduction=outvars[key][j]
        elif outvars[key][j][kpindex]==kpointv and outvars[key][j][bdindex]==bandv:
                valence = outvars[key][j]
    # Then the gap can be calculated
    array[i][1] = conduction[e0index]+conduction[gwindex]-(valence[e0index]+valence[gwindex])

if text:
    filename = folder+'_'+var+'.dat'
    header = +var+'('+str(unit)+'), gap'
    np.savetxt(filename,array,delimiter='\t',header=header)
    print(filename)

plt.plot(array[:,0],array[:,1],'o-')
plt.xlabel(var+' ('+unit+')')
plt.ylabel('E_gw = E_lda + \Delta E')
plt.show()
#plt.savefig(folder+'_'+var+'.png')

# Plot all of the different GW bandstructures in the same plot
#ya = YamboAnalyser(folder)
#ya.plot_gw('qp',cols=(lambda x: x[2],lambda x: x[3]+x[4]))




import time
from functools import wraps

# decarator to measure number of calls and wall time in seconds of function
"""
@func_profile
def abc(xxx):
    np.log(np.sin(xxx)+1j+np.cos(xxx)+xxx**2)

x = np.linspace(0,10,1000000)

for i in range(10):
    abc(x)

# number of calls to abc
print(abc.call_count)

# total time spent in abc
print(abc.total_time)

"""
def func_profile(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        
        # Increment call count (initialize if first call)
        if not hasattr(wrapper, "call_count"):
            wrapper.call_count = 0
        wrapper.call_count += 1

        result = func(*args, **kwargs)  # Call original function

        # Accumulate time (initialize if first call)
        if not hasattr(wrapper, "total_time"):
            wrapper.total_time = 0.0
        wrapper.total_time += time.perf_counter() - start_time

        return result
    return wrapper



# Copyright (c) 2023, Claudio Attaccalite
# All rights reserved.
#
# This file is part of the yambopy project
#
from netCDF4 import Dataset
from yambopy.units import ha2ev,fs2aut,speed_of_light
import numpy as np
import sys
import os
#
# This class reads all data from the ndb.Nonlinear database
# and its fragment ndb.Nonlinear_fragment_xxx
# 
# Time series is stored in IO_TIME_points variable
# while external fields, current and polarization of the different runs 
# are stored in Efield[x],Polarization[x],Current[x]
# It is a single run just use Efield[0],Current[0] etc...
#
class YamboNLDB(object):
    """
    Open the NL databases and store it in a NLDB class.
    """

    def __init__(self,folder='.',calc='SAVE',nl_db='ndb.Nonlinear'):
        # Find path with RT data
        self.nl_path = '%s/%s/%s'%(folder,calc,nl_db)
        self.calc=calc
        try:
            data_obs= Dataset(self.nl_path)
        except:
            raise ValueError("Error reading NONLINEAR database at %s"%self.nl_path)

        self.read_observables(data_obs)


        data_obs.close()

    def read_Efield(self,database,RT_step,n):
         efield={}
         efield["name"]       =database.variables['Field_Name_'+str(n)][...].tobytes().decode().strip()
         efield["versor"]     =database.variables['Field_Versor_'+str(n)][:].astype(np.double)
         efield["intensity"]  =database.variables['Field_Intensity_'+str(n)][0].astype(np.double)
         try:
            efield["damping"]    =database.variables['Field_FWHM_'+str(n)][0].astype(np.double)
         except:      
            efield["damping"]    =database.variables['Field_Damping_'+str(n)][0].astype(np.double)
         try:
             efield["freq_range"] =database.variables['Field_Freq_range_'+str(n)][:].astype(np.double)
         except:
             efield["freq_range"] =database.variables['Field_Freq_'+str(n)][:].astype(np.double)
         try:
             efield["freq_steps"] =database.variables['Field_Freq_steps_'+str(n)][:].astype(np.double)
         except:
             efield["freq_steps"] =1.0
         try:
             efield["freq_step"]  =database.variables['Field_Freq_step_'+str(n)][0].astype(np.double)
         except:
             efield["freq_step"]  =0.0
         efield["initial_time"]  =database.variables['Field_Initial_time_'+str(n)][0].astype(np.double)
         try:
             efield["peak"]  =database.variables['Field_peak_'+str(n)][0].astype(np.double)
         except:
             efield["peak"]  =10.0
         #
         # set t_initial according to Yambo 
         #
         efield["initial_indx"] =max(round(efield["initial_time"]/RT_step)+1,2)
         efield["initial_time"] =(efield["initial_indx"]-1)*RT_step
         #
         # define the field amplitude
         #
         efield["amplitude"]    =np.sqrt(efield["intensity"]*4.0*np.pi/speed_of_light)

         return efield

    def read_observables(self,database):
        """
        Read all data from the database
        """
        self.Gauge          = database.variables['GAUGE'][...].tobytes().decode().strip()
        self.NE_steps       = database.variables['NE_steps'][0].astype('int')
        self.RT_step        = database.variables['RT_step'][0].astype(np.double)
        self.n_frequencies  = database.variables['n_frequencies'][0].astype('int')

        try:
            self.n_angles       = database.variables['n_angles'][0].astype('int')
        except:
            self.n_angles   = 0
        try:
            self.NL_initial_versor = database.variables['NL_initial_versor'][:].astype(np.double)
        except:
            self.NL_initial_versor = [0.0, 0.0, 0.0] 

        self.NL_damping     = database.variables['NL_damping'][0].astype(np.double)
        self.RT_bands       = database.variables['RT_bands'][:].astype('int')
        self.NL_er          = database.variables['NL_er'][:].astype(np.double)
        self.l_force_SndOrd = database.variables['l_force_SndOrd'][0].astype('bool')
        self.l_use_DIPOLES  = database.variables['l_use_DIPOLES'][0].astype('bool')
        try:
            self.l_eval_CURRENT = database.variables['l_eval_CURRENT'][0].astype('bool')
        except:
            self.l_eval_CURRENT = False
        self.QP_ng_SH       = database.variables['QP_ng_SH'][0].astype('int')
        self.QP_ng_Sx       = database.variables['QP_ng_Sx'][0].astype('int')
        self.RAD_LifeTime   = database.variables['RAD_LifeTime'][0].astype(np.double)
        self.Integrator     = database.variables['Integrator'][...].tobytes().decode().strip()
        self.Correlation    = database.variables['Correlation'][...].tobytes().decode().strip()
        #
        # Time variables
        #
        self.IO_TIME_N_points  = database.variables['IO_TIME_N_points'][0].astype('int')
        self.IO_TIME_LAST_POINT= database.variables['IO_TIME_LAST_POINT'][0].astype('int')
        self.IO_TIME_points    = database.variables['IO_TIME_points'][:].astype(np.double)
        #
        # External fields
        # 
        self.Efield_general=[]
        self.N_ext_fields=0
        for n in range(1,4):
            try:
                efield=self.read_Efield(database,self.RT_step,n)
            except:
                print("Field %d not found" % n)
            else:
                self.Efield_general.append(efield.copy())
                self.N_ext_fields=self.N_ext_fields+1

        #
        # Read polarization and currect files 
        #
        self.Polarization=[]
        self.Current     =[]
        self.E_ext       =[]
        self.E_tot       =[]
        self.E_ks        =[]
        self.Efield      =[] # Store the first external field for each run at different frequencies
        self.Efield2     =[]
        #
        if self.n_angles!=0:
            self.n_runs=self.n_angles
        if self.n_frequencies!=0:
            self.n_runs=self.n_frequencies
        if (self.n_angles!=0 and self.n_frequencies!=0):
            print("Error both n_angles and n_frequencies !=0 ")
            sys.exit(0)
        if (self.n_angles==0 and self.n_frequencies==0):
            self.n_runs=1

        #
        for f in range(self.n_runs):
            try:
                data_p_and_j= Dataset(self.nl_path+"_fragment_"+str(f+1))
            except:
                print("Error reading database: %s" % self.nl_path+"_fragment_"+str(f+1))
                continue
            pol  = data_p_and_j.variables['NL_P_freq_'+str(f+1).zfill(4)][:,:].astype(np.double)
            curr = data_p_and_j.variables['NL_J_freq_'+str(f+1).zfill(4)][:,:].astype(np.double)
            e_ext= data_p_and_j.variables['E_ext_freq_'+str(f+1).zfill(4)][:,:,:].astype(np.double)
            e_ext_c=e_ext[:,:,0]+1j*e_ext[:,:,1]
            e_tot= data_p_and_j.variables['E_tot_freq_'+str(f+1).zfill(4)][:,:,:].astype(np.double)
            e_tot_c=e_tot[:,:,0]+1j*e_tot[:,:,1]
            e_ks = data_p_and_j.variables['E_ks_freq_'+str(f+1).zfill(4)][:,:,:].astype(np.double)
            e_ks_c=e_ks[:,:,0]+1j*e_ks[:,:,1]

            self.Polarization.append(pol.copy())
            self.Current.append(curr.copy())
            self.E_ext.append(e_ext_c.copy())
            self.E_tot.append(e_tot_c.copy())
            self.E_ks.append(e_ks_c.copy())

            # Read only the first field for SHG
            # I don't need it in the pump-probe configuration
            efield=self.read_Efield(data_p_and_j,self.RT_step,1)
            try:
                efield2=self.read_Efield(data_p_and_j,self.RT_step,2)
            except:
                efield2=efield
            self.Efield.append(efield.copy())
            self.Efield2.append(efield2.copy())

    def __str__(self):
        """
        Print all info of the database
        """
        s="\n * * * ndb.Nonlinear db data * * * \n\n"
        s+="Gauge         : "+str(self.Gauge)+"\n"
        s+="NE_steps      : "+str(self.NE_steps)+"\n"
        s+="RT_step       : "+str(self.RT_step/fs2aut)+" [fs] \n"
        s+="n_frequencies : "+str(self.n_frequencies)+"\n"   
        s+="n_angles      : "+str(self.n_angles)+"\n"   
        s+="NL_initial_versor   : "+str(self.NL_initial_versor)+"\n"   
        s+="NL_damping    : "+str(self.NL_damping*ha2ev)+"\n"
        s+="RT_bands      : "+str(self.RT_bands)+"\n"
        s+="NL_er         : "+str(self.NL_er*ha2ev)+" [eV] \n"
        s+="Sencond Order : "+str(self.l_force_SndOrd)+"\n" 
        s+="Use Dipoles   : "+str(self.l_use_DIPOLES)+"\n"
        s+="QP_ng_SH      : "+str(self.QP_ng_SH)+"\n"
        s+="QP_ng_Sx      : "+str(self.QP_ng_Sx)+"\n"  
        s+="RAD_LifeTime  : "+str(self.RAD_LifeTime/fs2aut)+" [fs] \n" 
        s+="Integrator    : "+str(self.Integrator)+"\n"
        s+="Correlation   : "+str(self.Correlation)+"\n"
        for efield in self.Efield:
            if efield["name"] == "none":
                continue
            s+="\nEfield name         : "+str(efield["name"])+"\n"
            s+="Efield versor       : "+str(efield["versor"])+"\n"  
            s+="Efield Intesity     : "+str(efield["intensity"])+"\n"
            s+="Efield Damping      : "+str(efield["damping"])+"\n"
            s+="Efield Freq range   : "+str(efield["freq_range"]*ha2ev)+" [ev] \n"
            s+="Efield Initial time : "+str(efield["initial_time"]/fs2aut)+" [fs] \n"
        return s




# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
import unittest
import os
import shutil as sh
from yambopy.analyse import YamboAnalyser
from yambopy.io.outputfile import YamboOut
from qepy.lattice import Path

test_path = os.path.join(os.path.dirname(__file__),'..','data','refs','gw_conv')

class TestYamboAnalyse(unittest.TestCase):
    def setUp(self):
        """ Read the yambo GW output files
        """
        if os.path.isdir('gw_conv'): sh.rmtree('gw_conv')
        sh.copytree(test_path,'gw_conv')
        for dirpath,dirnames,filenames in os.walk('gw_conv'):
            if YamboOut.has_output(dirpath):
                y = YamboOut(dirpath,save_folder='gw_conv')
                y.pack()

    def test_yamboanalyse_gw_si(self):
        """ Analyse the yambo GW .json output files
        """
        y = YamboAnalyser('gw_conv')
        netcdf_files = y.get_files_type('netcdf_gw')
        keys = sorted(netcdf_files.keys())
        assert keys == ['BndsRnXp_1_20', 'BndsRnXp_1_30', 'FFTGvecs_00010', 'FFTGvecs_00015', 
                        'NGsBlkXp_00002', 'NGsBlkXp_00005', 'reference']

        netcdf_files = y.get_files_type('netcdf_gw','FFTGvecs')
        keys = sorted(netcdf_files.keys())
        assert keys == ['FFTGvecs_00010', 'FFTGvecs_00015']

        netcdf_files = y.get_files_type('netcdf_gw',('FFTGvecs','reference'))
        keys = sorted(netcdf_files.keys())
        assert keys == ['FFTGvecs_00010', 'FFTGvecs_00015','reference']

        #test getting data
        ks_bands,qp_bands = y.get_bands(tags='reference')
        ks_bands.plot(show=False)
        print(ks_bands)
        qp_bands.plot(show=False)
        print(qp_bands)

        #test get_path    
        path = Path([[[1.0,1.0,1.0],'G'],
                     [[0.0,0.5,0.5],'X'],
                     [[0.0,0.0,0.0],'G'],
                     [[0.5,0.0,0.0],'L']], [20,20,20])
        ks_bands_path,qp_bands_path = y.get_bands(tags='FFTGvecs',path_kpoints=path,type_calc=('gw'))
        qp_bands_path.plot(show=False) 

    def tearDown(self):
        sh.rmtree('gw_conv') 
    
if __name__ == '__main__':
    unittest.main()



"""
Tutorial for YamboExcitonDB.

Plotting BSE optical spectra

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='BSE_saves/YAMBO_saves'
bse_path ='BSE_saves/BSE_databases'
from yambopy import YamboLatticeDB,YamboExcitonDB
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":

    Absorption_Plot = True

    # Customly chosen Q-point
    iQ=1 # 1-> Gamma point, i.e., optical absorption limit

    #                    #
    # Start Yambopy part #
    #                    #

    # Create "lattice" object by reading the ns.db1 database inside the yambo SAVE
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')

    # Read exciton data at Q=iQ
    yexc = YamboExcitonDB.from_db_file(ylat,filename=bse_path+'/ndb.BS_diago_Q%d'%iQ)

    #
    # Spectrum of the dielectric function (optical absorption at q=0)
    #
    # In order to customise your plot, try to test with different values of: 
    #
    #  * emin, emax, estep
    #  * broad
    #  * nexcitons
    #  * iQ
    #
    if Absorption_Plot:
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        ax.set_xlim(0,7)

        w, chi = yexc.plot_chi_ax(ax,nexcitons='all',emin=0,emax=7,estep=0.005,broad=0.04)

        ax.set_ylim(0,np.max(chi.imag)*1.1)
        plt.show()

    # In order to get the eps(w) data we use the function get_chi
    w, eps = yexc.get_chi(emin=0,nexcitons='all',emax=7,estep=0.005,broad=0.04)
    print('Epsilon array: ',eps.shape, eps.dtype)

    #                  #
    # End Yambopy part #
    #                  #



import numpy as np
from yambopy.units import amu2ry

def descreen_el_ph(pseudoscreened_elph,ph_energies,ph_modes,Z,Zval,masses=None):
    """
    This function takes pseudo-screened el-ph matrix elements and "descreens" them returning
    the bare el-ph matrix elements under the approximation that the core static inverse
    dielectric function is proportional to Zval/Z.

    This is an approximate procedure from Eq. (XXX) of REF. (XXX).

    The phonon modes must be the orthonormal phonon modes coming from the diagonalization
    of the dynamical matrix, not the atomic-mass weighted phonon modes which are commonly
    used. In case the latter modes (not normalized) are the only ones available, the atomic
    masses in amu units need to be given as argument.

    Inputs:
    - pseudoscreened_elph: pseudoscreened electron-phonon matrix elements (numpy array)
                           units: same as ph_energies
                           format: [iq][ik][il][is][ib1][ib2]
    - ph_energies: phonon energies (numpy array)
                   units: same as pseudoscreened_elph
                   format: [iq][il]
    - ph_modes: phonon modes (numpy array)
                format: [iq][il][iat][ix]
    - Z: ordered list of atomic numbers (list)
                format: [iat]
    - Zval: ordered list of valence electrons (list)
            format: [iat]
    - masses [optional]: atomic masses (list)
                         units: amu
                         format: [iat]

    NB: ensure that the energy units are consistent between ph_energies and pseudoscreened_elph
        (i.e., hartree, rydberg, etc | QE uses RYDBERG | Yambo uses HARTREE)
    NB: if the masses array is provided, then the eigenvectors will be weighted by the masses
        to restore orthonormality. Remember to give the masses in AMU units.

    Example usage reading the data from LetzElPh output

    ```

    from yambopy import LetzElphElectronPhononDB
    from yambopy.units import ha2ev
    ry2ev = ha2ev/2.              # lelphc uses RYDBERG units
    Z = [42,16,16]                # atomic_numbers (hardcoded here, read with pwin or latticedb)
    Zval = [14,6,6]               # Taken from pseudopotentials, read with qepy utility
    M_amu = [95.95,32.065,32.065] # atomic masses in amu (hardcoded here, read with qepy/yambopy)

    ygpp  = LetzElphElectronPhononDB('PATH/ndb.elph') # dbs generated with "kernel=bare" in lelphc input

    bare_gkkp = descreen_el_ph(ygpp.gkkp,ygpp.ph_energies/ry2ev,ygpp.ph_eigenvectors,Z,Zval,masses=M_amu)

    ```
    """
    na = np.newaxis
    Z = np.array(Z)
    Zval = np.array(Zval)
    if masses is not None:
        masses = amu2ry*np.array(masses)
        ph_modes = ph_modes*np.sqrt( masses[na,na,:,na] )

    # Compute mode overlap matrix weighted by core charge
    L = np.einsum('a,qmax,qnax->qmn',Z/Zval,np.conj(ph_modes),ph_modes,optimize=True)
    # Energy-weighted overlap matrix
    Lw = np.conj(L)*np.sqrt(ph_energies[:,:,na]/ph_energies[:,na,:])
    # Bare matrix elements
    bare_elph =  np.einsum('qmn,qkmsvc->qknsvc',Lw,pseudoscreened_elph,optimize=True)
    
    return bare_elph



#
# Author: Alejandro Molina-Sanchez
#
# Example of YamboQPDB Class 
#
from qepy import *
from yambopy import *
import matplotlib.pyplot as plt


# Define path in reduced coordinates using Class Path
npoints = 10
path = Path([ [[  0.0,  0.0,  0.0],'$\Gamma$'],
              [[  0.5,  0.0,  0.0],'M'],
              [[1./3.,1./3.,  0.0],'K'],
              [[  0.0,  0.0,  0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

# Read Lattice information from SAVE
lat  = YamboLatticeDB.from_db_file(filename='gw/SAVE/ns.db1')
# Read QP database
ydb  = YamboQPDB.from_db(filename='ndb.QP',folder='gw/yambo')


# 1. Find scissor operator for valence and conduction bands

fig = plt.figure(figsize=(6,4))
ax  = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])
ax.set_xlabel('$E_{KS}$')
ax.set_ylabel('$E_{GW}$')

n_top_vb = 4
ydb.plot_scissor_ax(ax,n_top_vb)

plt.show()

# 2. Plot of KS and QP eigenvalues NOT interpolated along the path
n_top_vb = 3
ks_bs_0, qp_bs_0 = ydb.get_bs_path(lat,path)

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs_0.set_fermi(n_top_vb)
qp_bs_0.set_fermi(n_top_vb)

ks_bs_0.plot_ax(ax,legend=True,color_bands='r',label='KS')
qp_bs_0.plot_ax(ax,legend=True,color_bands='b',label='QP-GW')

plt.show()

# 3. Interpolation of KS and QP eigenvalues

ks_bs, qp_bs = ydb.interpolate(lat,path,what='QP+KS',lpratio=20)

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs.set_fermi(n_top_vb)
qp_bs.set_fermi(n_top_vb)

ks_bs.plot_ax(ax,legend=True,color_bands='r',label='KS')
qp_bs.plot_ax(ax,legend=True,color_bands='b',label='QP-GW')

plt.show()

# 4. Comparison of not-interpolaed and  interpolated eigenvalues

fig = plt.figure(figsize=(4,5))
ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])

ks_bs_0.plot_ax(ax,legend=True,color_bands='r',label='KS')
qp_bs_0.plot_ax(ax,legend=True,color_bands='b',label='QP-GW')
ks_bs.plot_ax(ax,legend=True,color_bands='g',label='KS')
qp_bs.plot_ax(ax,legend=True,color_bands='k',label='QP-GW')

plt.show()



import matplotlib.pyplot as plt
from qepy import Spin_texture

# Figure features
fig = plt.figure(figsize=(7,7))
ax = fig.add_axes( [ 0.15, 0.1, 0.8, 0.8 ])

# DB definitions
prefix = 'bn'
path = '.'
folder_spin = 'spin_bands'
prefix_spin = 'spin'

# Initializing Spin_texture class
Spin_tex = Spin_texture(prefix=prefix,path=path,folder_spin=folder_spin,prefix_spin=prefix_spin)

# Load spin projection values (i = 1,2,3 = x,y,z)
Spin_x = Spin_tex.load_spin_data(1)

print(Spin_x)

# Plot spin_texture (mode = "raw", "interpolated", "arrow")
Spin_tex.plot_spin_texture(ax,limfactor=0.8,mode="raw",nband=5)

plt.show()



